# Author(s): Kristo Mela
# This source code is licensed under the MIT license. See LICENSE in the repository root directory.
# Copyright 2022 Kristo Mela
# -*- coding: utf-8 -*-
"""
Created on Thu Oct 17 19:53:07 2019

Composite beam sections

@author: kmela
"""

import math
import numpy as np

import matplotlib.pyplot as plt
import matplotlib.patches as patches

from metku.eurocodes.en1993 import en1993_1_1, en1993_1_5, constants
from metku.eurocodes.en1994 import en1994_1_1
    

class ConcreteSlab:
    """ Simple concrete slab with rectangular cross-section
        
    
    """
    
    def __init__(self,thickness,material,reinforcement=None):
        """ Constructor
            input:
                thickness [mm] thickness of slab
                material .. Concrete class material
        """
        
        self.hc = thickness
        self.material = material
        
    @property
    def h(self):
        """ Total height of the slab """
        return self.hc
    
    @property
    def h0(self):
        """ Notional size of slab thickness, for drying (EN 1992-1-1, Annex B) """
        return 2*self.hc
        
    def weight(self):
        """ Weight of the slab """
        return (self.material.density+1)*self.hc*1e-3
        
    def draw(self,axes=None,y0=0,width=100):
        """ Draw slab """
        if axes is None:
            fig, ax = plt.subplots(1)
        else:
            ax = axes
        
        
        
        slab = patches.Rectangle((-0.5*width, y0), width=width,
                                       height=self.hc, hatch='\\')
        
        ax.add_patch(slab)
        
        
    
class CompositeSlab:
    """ Simple composite slab with rectangular cross-section
        
    
    """
    
    def __init__(self,h_total,concrete,h_steel):
        """ Constructor
            input:
                thickness [mm] thickness of slab
                material .. Concrete class material
        """
        
        self.hc = h_total-h_steel
        self.material = concrete
        self.ha = h_steel
        
    @property
    def h(self):
        """ Total height of the slab """
        return self.hc + self.ha
    
    @property
    def h0(self):
        """ Notional size of slab thickness, for drying (EN 1992-1-1, Annex B) """
        return 2*self.hc
    
    def concrete_weight(self):
        """ Weight of the concrete """
        return (self.material.density+1)*self.h*1e-3
    
    def draw(self,axes=None,y0=0,width=100):
        """ Draw slab """
        if axes is None:
            fig, ax = plt.subplots(1)
        else:
            ax = axes
        
        
        
        slab_steel = patches.Rectangle((-0.5*width, y0), width=width,
                                       height=self.ha, fill=False)
        
        slab_concrete = patches.Rectangle((-0.5*width, y0+self.ha), width=width,
                                       height=self.hc, hatch='\\')
        ax.add_patch(slab_steel)
        ax.add_patch(slab_concrete)
        

class CompositeIBeam:
    """ Class for composite beams with an I section as
        steel part and a slab (composite or concrete) on top
    """
    
    def __init__(self,length,steel_part,slab,studs=None,RH=50):
        """ Constructor
            Input:
                length .. length of the beam
                steel_part: ISection class profile or WI profile steel section
                slab: concrete or composite slab
                studs: shear studs
                RH: relative humidity of concrete
                
            parameters
                length .. length of the beam
                steel .. steel profile
                slab .. slab
                studs .. shear connector, of class ShearStud
                st .. stud spacing
                sf .. transverse reinforcement spacing
                
        """
        
        self.length = length
        self.steel = steel_part
        self.slab = slab
        self.slab.material.RH = RH
        self.studs = studs
        self.st = 150
        self.sf = 150
        self.b0 = 0                 
        self.t0 = 28 # Time for creep (days)         
        
    def creep_coefficient(self,t0=None):
        
        if t0 is None:
            t0 = self.t0
            
        return self.slab.material.creep_coefficient(self.slab.h0,1e8,t0)
    
    @property
    def n0(self):
        """ Ratio of Young's modulae of steel and concrete (short term) """
        return self.Ea/self.Ecm
    
    @property
    def nC(self):
        """ Ratio of Young's modulae of steel and concrete (creep)"""
        return self.Ea/self.Ecc
    
    @property
    def nS(self):
        """ Ratio of Young's modulae of steel and concrete (shrinkage)"""
        return self.Ea/self.Ecs
        
    @property 
    def beff(self):
        """ Effective width of concrete flange """
        return self.b0 + 2.0*self.length/8.0
        
    @property
    def yc(self):
        """ Distance of centroid of concrete slab from the top
            of the composite section
        """
        return 0.5*self.slab.hc
    
    @property
    def ya(self):
        """ Distance of centroid of steel part from the top
            of the composite section
        """
        return self.slab.h + 0.5*self.steel.h
    
    @property
    def Aa(self):
        """ Area of steel part """
        return self.steel.A
    
    @property
    def Ia(self):
        """ Second moment of area of steel part with respect to
            the major axis at centroid of steel part
        """
        return self.steel.I[0]
    
    @property
    def Ea(self):
        """ Young's modulus of steel part """
        return self.steel.E
    
    @property
    def EAa(self):
        """ Axial stiffness of steel part """
        return self.Ea*self.Aa
    
    @property
    def EIa(self):
        """ Bending stiffness of steel part 
            with respect to the neutral axis of steel part
        """
        return self.Ea*self.Ia
    
    @property
    def Ecm(self):
        """ Young's modulus of concrete """
        return self.slab.material.Ecm
    
    @property
    def Ecc(self):
        """ Young's modulus of concrete with creep effects """        
        phi = self.creep_coefficient()        
        return self.Ecm/(1+1.1*phi)
    
    @property
    def Ecs(self):
        """ Young's modulus of concrete with shrinkage effects """        
        phi = self.creep_coefficient(t0=1)        
        return self.Ecm/(1+0.55*phi)
    
    @property
    def EIc(self):
        """ Bending stiffness of concrete part 
            with respect to the neutral axis of concrete part
        """
        return self.Ecm*self.Ic
    
    @property
    def Ac(self):
        """ Area of concrete part """
        return self.slab.hc*self.beff
    
    @property
    def Ic(self):
        """ Second moment of area of concrete part with respect to
            the major axis at centroid of concrete part
        """
        return 1/12*self.beff*self.slab.hc**3
    
    @property
    def EAc(self):
        """ Axial stiffness of concrete """
        return self.Ecm*self.Ac
        
    @property
    def EAcc(self):
        """ Axial stiffness of concrete in creep"""
        return self.Ecc*self.Ac
    
    @property
    def EIcc(self):
        """ Bending stiffness of concrete part with creep effects
            with respect to the neutral axis of concrete part
        """
        return self.Ecc*self.Ic
    
    def centroid(self,long_term=None):
        """ Centroid of the composite section 
            Assume full interaction
        """        
        
        if long_term is None:
            yN = (self.EAa*self.ya + self.EAc*self.yc)/self.EAcom()
        elif long_term == "creep":
            yN = (self.EAa*self.ya + self.EAcc*self.yc)/self.EAcom("creep")
                        
        return yN
    
    def EAcom(self,long_term=None):
        """ axial stiffness composite section """
        if long_term is None:
            EAcom = self.EAa + self.EAc
        elif long_term == "creep":            
            EAcom = self.EAa + self.EAcc
        
        return EAcom


    def ya_com(self,long_term=None):
        """ distance of centroid of steel part from the neutral
            axis of the composite section
        """
        
        return abs(self.ya-self.centroid(long_term))
    
    def yc_com(self,long_term=None):
        """ distance of centroid of concrete part from the neutral
            axis of the composite section
        """
        
        return abs(self.yc-self.centroid(long_term))

    def EIcom(self,long_term=None):
        """ sending stiffness of composite section """        
        if long_term is None:
            EIcom = self.Ea*(self.Ia+self.ya_com()**2*self.Aa) + self.Ecm*(self.Ic+self.yc_com()**2*self.Ac)
        elif long_term == "creep":
            EIcom = self.Ea*(self.Ia+self.ya_com(long_term)**2*self.Aa) + self.Ecc*(self.Ic+self.yc_com(long_term)**2*self.Ac)
        
        return EIcom

    def distance_centroids(self):
        """ distance between centroids of steel and concrete parts """
        return self.ya - self.yc
    
    def composite_coefficient(self):
        """ Coefficint alpha that indicates the degree of composite action 
            TRY/BY, pp. 24        
        """
        a = self.distance_centroids()
        return a**2*self.EAa*self.EAc/(self.EAa+self.EAc)/(self.EIa + self.EIc)
    
    def Ra(self):
        """ Full resultant of steel part """
        
        return self.steel.fy*self.Aa
    
    def Rc(self):
        """ Full resultant of concrete part """
        
        return 0.85*self.slab.material.fcd()*self.Ac
    
    def Rf(self):
        """ Resultant of top flange of the steel beam """
        
        Af = self.steel.tf * self.steel.b        
        
        return self.steel.fy*Af
    
    def Rw(self):
        """ Resultant of web of the steel beam """
        
        Aw = self.steel.hw * self.steel.tw        
        
        return self.steel.fy*Aw
    
    def plastic_neutral_axis(self):
        """ Location of the plastic neutral axis
            from the top of the concrete slab
        """
        Ra = self.Ra()
        Rc = self.Rc()
        
        hc = self.slab.hc
        
        if Ra < Rc:
            ec0 = Ra/Rc*hc
        elif Rc < Ra-2*self.Rf():
            hl = self.slab.h
            ec0 = 0.5*(Ra-Rc)/self.steel.b/self.steel.fy + self.slab.h
        else:
            Rf = self.Rf()
            Rw = self.Rw()
            hw = self.steel.hw
            hl = self.slab.h
            tf = self.steel.tf
            
            ec0 = 0.5*(Ra-Rc-2*Rf)/Rw*hw+hl+tf
            
        return ec0
        
    
    def MplRd(self,verb=False):
        """ Plastic moment resistance based on full shear connection """
        
        Ra = self.Ra()
        Rc = self.Rc()
        
        ei = self.distance_centroids()
        hc = self.slab.hc
        if Ra < Rc:
            """ Neutral axis in concrete part """            
            ec0 = Ra/Rc*hc
            MplRd = Ra*(ei+0.5*hc-0.5*ec0)
            if verb:
                print("Plastic neutral axis in concrete:")
                print("Ra = {0:4.2f} kN".format(Ra*1e-3))
                print("Rc = {0:4.2f} kN".format(Rc*1e-3))
                print("ec0 = {0:4.2f} [mm]".format(ec0))
        elif Rc < Ra-2*self.Rf():
            """ Neutral axis in top flange of steel beam """
            hl = self.slab.h
            ec0 = 0.5*(Ra-Rc)/self.steel.b/self.steel.fy + self.slab.h
            MplRd = Ra*(ei-hl+0.5*hc) + Rc*(hl-0.5*hc) - 0.25*(Ra-Rc)**2/self.Rf()*self.steel.tf
        else:
            """ Neutral axis in the web of the steel part """
            Rf = self.Rf()
            Rw = self.Rw()
            hw = self.steel.hw
            hl = self.slab.h
            tf = self.steel.tf
            
            ec0 = 0.5*(Ra-Rc-2*Rf)/Rw*hw+hl+tf
            
            MplRd = Ra*(ei-hl-tf+0.5*hc) + Rc*(hl+tf-0.5*hc) + Rf*tf - 0.25*(Ra-Rc-2*Rf)**2/Rw*hw
                    
        if verb:
            print("MplRd = {0:4.2f} [kNm]".format(MplRd*1e-6))
        
        return MplRd
    
    def VRd(self):
        """ Shear resistance
            This equals shear resistance of the steel part
        """
        
        return self.steel.shear_force_resistance()
    
    def draw(self):
        """ Draw the profile             
        """
        
        fig, ax = plt.subplots(1)

        beff = self.beff

        self.steel.draw(axes=ax)        
                    
        
        
        self.slab.draw(axes=ax,y0=0.5*self.steel.h,width=beff)    
            
        yel = 0.5*self.steel.h + self.slab.h-self.centroid()
        ypl = 0.5*self.steel.h + self.slab.h-self.plastic_neutral_axis()      
        
                        
        plt.plot(0.0, yel, 'or')
        plt.plot(0.0, ypl, 'db')
    
        """
        tflange =  mlines.Line2D(top_x_data,top_y_data,linewidth=lw)
        ax.add_line(tflange)
        
        
        # Top chamfers
        tr_corner = patches.Arc((0.5*tw+r,0.5*hw),2*r,2*r,0,90,180)
        tl_corner = patches.Arc((-0.5*tw-r,0.5*hw),2*r,2*r,0,0,90)
        
        ax.add_patch(tr_corner)
        ax.add_patch(tl_corner)
        
        # Draw web
        xweb_l = [-0.5*tw,-0.5*tw]
        yweb = [-0.5*hw,0.5*hw]
        
        if theta is not 0:
            web_rot = [R.dot(np.array([x,y])) for x, y in zip(xweb_l,yweb)]
            xweb_l = [x[0] for x in web_rot]
            yweb_l = [x[1] for x in web_rot]
        else:
            yweb_l = yweb
        
        lweb = mlines.Line2D(xweb_l,yweb_l,linewidth=lw)
        ax.add_line(lweb)
        
        xweb_r = [0.5*tw,0.5*tw]
        
        rweb = mlines.Line2D(xweb_r,yweb,linewidth=lw)
        ax.add_line(rweb)
        """
        
        """
        verts = [(0,0),
                 (1,0),
                 (0,1)]
        
        codes = [mpath.Path.]
        
        Iprof_path = mpath.Path(verts,codes)
        
        
        Iprof = patches.PathPatch(Iprof_path,facecolor="none",lw=1.5)
        ax.add_patch(Iprof)
        """
        
        #ax.add_patch(bot_flange)
        #ax.add_patch(web)
        #ax.add_patch(top_flange)
        #ax.set_xlim(-0.5*b, 0.5*b)
        #ax.set_ylim(-0.5*h, 0.5*h)
         
        ax.set_xlim(-0.5*beff, 0.5*beff)
        ax.set_ylim(-0.5*self.steel.h, 0.5*self.steel.h+self.slab.h)
        
        ax.set_aspect('equal')


if __name__ == '__main__':
    
        
    from sections.steel.ISection import IPE
    from eurocodes.en1992 import en1992_1_1, constants


    """    
    steel = IPE(300)
    #slab = ConcreteSlab(140,constants.Concrete("C25/30"))
    slab = CompositeSlab(100,constants.Concrete("C25/30"),50)
    
    p = CompositeIBeam(12000,IPE(450),slab)
    p.draw()
    
    print(p.slab.material.creep_coefficient(50,227,t=1e8,t0=28))
    
    print("Effective width of concrete flange {0:4.2f} mm".format(p.beff))    
    print("Centroid of concrete {0:4.2f} mm".format(p.yc))
    print("Centroid of steel part {0:4.2f} mm".format(p.ya))
    print("Centroid {0:4.2f} mm".format(p.centroid()))
    print("Bending stiffness of steel part: {0:4.2f} MNm".format(p.EIa*1e-12))
    print("Bending stiffness of concrete part: ",p.EIc*1e-12)
    print("Distance between centroids [mm]: ", p.distance_centroids())
    print("Liittovaikutuskerroin: ", p.composite_coefficient())
    print("Bending stiffness of composite section: ",p.EIcom()*1e-12)
    print("Bending resistance: MplRd = {0:4.2f}".format(p.MplRd()*1e-6))
    print("Bending resistanc of steel: Mpl,a,Rd = {0:4.2f}".format(p.steel.plastic_bending_resistance()*1e-6))
    """


# Author(s): Kristo Mela
# This source code is licensed under the MIT license. See LICENSE in the repository root directory.
# Copyright 2022 Kristo Mela
# -*- coding: utf-8 -*-
""" Particle swarm optimization """

import operator
import random

import matplotlib.pyplot as plt
import numpy as np
from deap import base, creator, tools

from metku.optimization.solvers.optsolver import OptSolver


class PSO(OptSolver):
    """ Particle swarm optimization solver """
    def __init__(self,pop_size=10, c = [2.05,2.05], inertia=0.9, craziness=0.1, 
                 penalty=None,best_f = np.inf):
        """ Constructor
            :param pop_size: population size
            :c: [c1,c2] weighting factors
            :inertia: initial value for the inertia term
            :craziness: craziness probability            
        
        """

        self.xbest = None
        self.fbest = np.inf
        
        super().__init__()

        """ Toolbox contains basic evolutionary operators """
        self.toolbox = base.Toolbox()
        self.pop_size = pop_size
        self.mut_rate = mut_rate
        self.c = c
        self.inertia = w
        self.craziness = craziness
        self.plot = False
        self.best_f = best_f
        self.prev_best = best_f
        self.first_improvement = first_improvement
        self.best_ind = None
        if penalty is None:
            def penalty(constr_vals):
                pos_vals = np.clip(constr_vals, 0, 100)
                return 2e6 * sum(pos_vals ** 2)

            self.penalty = penalty
        else:
            self.penalty = penalty


        self.counter = 0

    def take_action(self):
        pass

    def step(self, action):
        pass

    def eval_individual(self, ind):
        """
        Evaluates one individual

        :param ind: 'individual', i.e. design variable values
        :return: objective value of optimization problem
        """
        self.problem.substitute_variables(ind)
        X = [round(var.value, 3) for var in self.problem.vars]
        
        """ Evaluate objective function """
        obj_val = self.problem.obj(X)
        """ Evaluate constraints """
        constr_vals = self.problem.eval_cons(X)
        """ Apply penalty on constraints """
        penalty_val = self.penalty(constr_vals)
        
        """ Fitness function equals objective function value + penalty """
        val = obj_val + penalty_val
        if self.plot:
            self.update_plot(self.fig, self.ax)


        # print("GAPS: ", [j.g1 for j in self.problem.structure.joints.values()])

        """ If a new best solution has been found, store it """
        if val < self.best_f:
            for i, x in enumerate(X):
                ind[i] = x
            self.best_ind = ind
            self.prev_best = self.best_f
            best_x = [var.value for var in self.problem.all_vars]
            self.best_x = best_x.copy()
            self.best_f = val
            self.counter = 0

        # NOTE! Need to return at least two values, hence the comma
        return val,  # <- DO NOT REMOVE THIS COMMA

    def create_fitness(self):

        """ Create Fitness class called FitnessMin 
            The weight -1.0 corresponds to minimization problem
        """
        creator.create("FitnessMin", base.Fitness, weights=(-1.0,))
        """ Create class Particle for a single particle in the swarm.
            Inherits the list class and has the FitnessMin as its fitness attribute.
            
            Also has the attributes
            speed .. current speed vector
            smin .. lower bound for speed
            smax .. upper bound for speed
            best .. reference to a copy of the best location the particle has been so far.
        """
        creator.create("Particle", list, fitness=creator.FitnessMax, speed=list, 
                       smin=None, smax=None, best=None)

    def attribute_generator(self, counter_list=[0]):
        """
        Generates one gene attribute
        :param counter_list: list used for counting how
                             many times function has been called
        :return:
        """
        idx = counter_list[0] % len(self.problem.vars)
        counter_list[0] = counter_list[0] + 1

        var = self.problem.vars[idx]

        """ Creates random integer value for index or integer variables """
        if isinstance(var, (IndexVariable, IntegerVariable)):
            val = np.random.randint(var.lb, var.ub)

        else:
            val = np.random.uniform(var.lb, var.ub)

        return val
    
    def generate(self, pmin, pmax, smin, smax):
        part = creator.Particle(random.uniform(pmin, pmax) for _ in range(size)) 
        part.speed = [random.uniform(smin, smax) for _ in range(size)]
        part.smin = smin
        part.smax = smax
        return part

    def register(self):
        """ Registers a set of functions to be used in the algorithm 
            the arguments for the register method are:
                1. alias for the function to be registered
                2. the actual method to registered
                3. arguments to be passed to the aliased method
        
        """
        self.toolbox.register("attribute", self.attribute_generator)

        self.toolbox.register("individual", tools.initRepeat,
                              creator.Individual,
                              self.toolbox.attribute, len(self.problem.vars))

        self.toolbox.register("particle", generate, size=2, pmin=-6, pmax=6, smin=-3, smax=3)
        self.toolbox.register("population", tools.initRepeat, list, toolbox.particle)
        self.toolbox.register("update", updateParticle, phi1=2.0, phi2=2.0)
        self.toolbox.register("evaluate", benchmarks.h1)

    def new_population(self, offspring):
        """
        Creates a new population from offsprings by shuffling them
        :param offspring: list of offsprings
        :return: new population
        """
        idx = np.random.randint(len(offspring)-1)
        self.problem.substitute_variables(self.best_x)
        values = [var.value for var in self.problem.vars]
        for i, val in enumerate(offspring[idx]):
            offspring[idx][i] = values[i]
        multiplier = self.pop_size // len(offspring) + 1
        new_population = list(offspring) * multiplier
        np.random.shuffle(new_population[:self.pop_size])

        if self.best_ind is not None:
            new_population.append(self.best_ind)

        return new_population


    def solve(self, problem, x0=None, maxiter=100, maxtime=-1, log=False,
              min_diff=1e-5, verb=False, plot=False):


        
        self.toolbox = base.Toolbox() # THIS COMMAND IS ALREADY IN THE CONSTRUCTOR!
        self.problem = problem
        """ Creates the fitness function 
            but the create_fitness method does not return anything!
        """
        self.create_fitness()
        self.register()

        if plot and not self.plot:
            fig = plt.figure()
            ax = fig.add_subplot(111)
            plt.ion()
            plt.show()
            self.plot = True
            self.fig = fig
            self.ax = ax

        """ Create population """
        pop = self.toolbox.population(n=self.pop_size)
        
        """ Evaluate fitness """
        fitnesses = list(map(self.toolbox.evaluate, pop))

        # CXPB  is the probability with which two individuals
        #       are crossed
        #
        # MUTPB is the probability for mutating an individual
        CXPB, MUTPB = self.cx_rate, self.mut_rate

        # Main loop
        it = 0
        prev_val = None
        while it < maxiter:
            it += 1
            problem.num_iters += 1
            print(f"Iteration {it}/{maxiter}")
            # Select the next generation individuals
            offspring = self.toolbox.select(pop)
            new_pop = self.new_population(offspring)

            # Clone the selected individuals
            offspring = list(map(self.toolbox.clone, new_pop))

            # Apply crossover and mutation on the offspring
            for child1, child2 in zip(offspring[::2], offspring[1::2]):
                # cross two individuals with probability CXPB
                if random.random() < CXPB:
                    self.toolbox.mate(child1, child2)

                    # fitness values of the children
                    # must be recalculated later
                    del child1.fitness.values
                    del child2.fitness.values
                for mutant in offspring[:-1]:
                    # mutate an individual with probability MUTPB
                    if random.random() < MUTPB:
                        self.toolbox.mutate(mutant)
                        del mutant.fitness.values

            # Evaluate the individuals with an invalid fitness
            invalid_ind = [ind for ind in offspring if
                           not ind.fitness.valid]
            fitnesses = map(self.toolbox.evaluate, invalid_ind)
            for ind, fit in zip(invalid_ind, fitnesses):
                ind.fitness.values = fit
            # The population is entirely replaced by the offspring
            pop[:] = offspring
            # Gather all the fitnesses in one list and print the stats
            fits = [ind.fitness.values[0] for ind in pop]
            length = len(pop)
            mean = sum(fits) / length
            sum2 = sum(x * x for x in fits)
            std = abs(sum2 / length - mean ** 2) ** 0.5

            if verb:
                print("VERB: ", self.best_x)
                print(f"Min fitness: {min(fits)} \n"
                      f"Obj: {problem.obj(self.best_x)} \n"
                      f"Max con: {max(problem.eval_cons(self.best_x)):.3f}  "
                      f"{problem.cons[np.argmax(problem.eval_cons(self.best_x))].name}")
            problem.substitute_variables(self.best_x)

            if plot:
                self.update_plot(self.fig, self.ax)

            if self.first_improvement and self.best_f < self.prev_best:
                return self.best_f, self.best_x

            if self.counter > (10 - 3):
                print(f"The result has not improved in the last 10 generations."
                      f"\n Returning current best values.")
                return self.best_f, self.best_x

            elif it > 0 and prev_val == min(fits):
                self.counter += 1

            else:
                prev_val = min(fits)

            if log:
                problem.num_iters += 1
                problem.fvals.append(self.best_f)
                problem.states.append(list(pop[np.argmin(fits)]))
                problem.gvals.append(max(self.constr_vals))
                self.fvals.append(self.best_f)
                self.xvals.append(self.best_x)

        # best_idx = np.argmin(fits)
        # xopt = pop[best_idx]
        # print("XOPT: ", xopt)
        # fopt = fits[best_idx]

        return self.best_f, self.best_x

class Population:
    """ Class for population (swarm) 
        
    """
    
    def __init__(self,pop_size):
        self.xbest = None
        self.fbest = np.inf
        self.parts = [0 for _ in range(pop_size)]
    
    

class Particle:

    def __init__(self, X, obj, parent):
        self.X = X
        self.obj = obj
        self.best_X = X

    @property
    def velocity(self):
        pass

    @property
    def best_f(self):
        return self.obj(self.best_X)

    @property
    def swarm_best_X(self):
        return self.parent.best_X

    def move(self):
        """
        Moves particle
        """
        self.X += self.velocity


if __name__ == '__main__':



# -*- coding: utf-8 -*-
# Copyright 2022 Kristo Mela
# This source code is licensed under the MIT license. See LICENSE in the repository root directory.
# Author(s): Kristo Mela
"""
Created on Tue Apr 19 10:14:37 2022

Raami functions for exporting to various programs

@author: kmela
"""

def write_elset(file,elset_name,els,instance=None):
    
    header = '*Elset, elset=' + elset_name
    
    if instance is None:
        header += '\n'
    else:
        header += ', instance=' + instance + '\n'
    
    file.write(header)
    i = 0
    while i < len(els):
        file.write(', '.join(str(r) for r in els[i:i+16]))
        file.write('\n')
        i += 16

def write_nset(file,nset_name,nodes,instance=None):
    
    header = '*Nset, nset=' + nset_name
    
    if instance is None:
        header += '\n'
    else:
        header += ', instance=' + instance + '\n'
    
    file.write(header)
    i = 0
    while i < len(nodes):
        file.write(', '.join(str(r) for r in nodes[i:i+16]))
        file.write('\n')
        i += 16

class AbaqusOptions:
    """ Storage for ABAQUS export options """
    
    def __init__(self,**kwargs):
        """
        

        Parameters
        ----------
        **kwargs : TYPE
            DESCRIPTION.

        Returns
        -------
        None.

        """
        
        self.x_monitor = 0
        self.n_monitored = 1
        self.mpc = []   # multi-point constraints 
        self.ecc_elements = [] # list of eccentricity elements
        self.top_gap_elements = [] # gap elements of the top chord
        self.bottom_gap_elements = [] # gap elements of bottom chord
        
        self.elsets = {} # *Elset groups
        
        self.load_elsets = {}
        self.load_nsets = {}
        
        # Individual members to be included as element sets
        self.included_members = []
        
        attr = self.__dict__.keys()
        
        for key, val in kwargs.items():
            if key in attr:
                setattr(self,key,val)


# Author(s): Kristo Mela
# This source code is licensed under the MIT license. See LICENSE in the repository root directory.
# Copyright 2022 Kristo Mela
# -*- coding: utf-8 -*-
"""
Created on Tue Dec 28 14:52:58 2021

Loads for Raami objects

@author: kmela
"""

import numpy as np
from copy import copy

import metku.framefem.framefem as fem

#from loadIDs import LoadIDs

LoadIDs = {'ULS': 0,
           'SLS_Characteristic': 1,
           'SLS_Frequent': 2,
           'SLS_Quasi_permanent': 3,
           'ACC': 4,
           'SELF-WEIGHT':100}

class LoadCase:
    """ Class for load cases. One load case represents a set of loads that
        are due to a single source. These loads can be of different type (PointLoad or LineLoad)
        and they can act on different parts of the structure, but they act
        simultaneously.
    """
    
    def __init__(self,load_id=0,loads=[]):
        """        

        Parameters
        ----------
        load_id : int, optional
            Unique load case identifier. The default is 0.
        loads : list, optional
            List of Load objects. The default is None.

        Returns
        -------
        None.

        """
        
        self.load_id = load_id
        self.loads = loads
    
    def __repr__(self):
        
        return f"Load case {self.load_id}. Total loads: {len(self.loads)}"
    
    @property
    def load_type(self):
        """ Returns the type of load """
        if len(self.loads) > 0:
            return self.loads[0].ltype
        else:
            return None
    
    def add(self,load):
        """ Adds a Load class object 'load' to the load case """
        load.load_id = self.load_id
        
        self.loads.append(load)

    def combine(self,comb_type='ULS',leading=True):
        """ Create factored loads for load combination of type comb_type """
        
        combined_loads = []
        
        for load in self.loads:
            comb_load = copy(load)
            comb_load.factored(comb_type,leading)
            combined_loads.append(comb_load)
        
        return combined_loads
            
        
class LoadCombination:
    """ Class for load combinations """
    
    
    def __init__(self,comb_id,comb_type='ULS',load_cases=[]):
        """
        
        Parameters
        ----------
        comb_id : float
            Unique identifier of the load combination.
        comb_type : string
            Type of combination. Can be 'ULS', 'SLS', or 'ACC'. This
            determines the combination factors
        load_cases : TYPE, optional
            DESCRIPTION. The default is [].

        Returns
        -------
        None.

        """
        
        self.comb_id = comb_id
        self.comb_type = comb_type
        self.load_cases = load_cases
        self.lead_variable_load = None
        
        # The first non-dead load is the leading variable load case.
        for lc in load_cases:
            if not lc.load_type == 'dead':
                self.lead_variable_load = lc
                break
        
        self.frame = None
    
    def combine_loads(self):
        """ Does the actual load combination:
            Each load in each load case is multiplied by its corresponding
            load combination factor that depends on the load type.
        """
        
        for lc in self.load_cases:
            # Do the load factorization for given load case
            # combo_loads is a list of factored loads.
            if lc == self.lead_variable_load:            
                combo_loads = lc.combine(self.comb_type,leading=True)
            else:
                combo_loads = lc.combine(self.comb_type,leading=False)
                
            for cl in combo_loads:
                cl.load_id = self.comb_id
                self.frame.add(cl)
        

class Load:
    """ General class for loads """

    def __init__(self, load_id, ltype='live', name='Load', f=1.0):
        """        

        Parameters
        ----------
        load_id : TYPE
            DESCRIPTION.
        ltype : string, optional
            Load type. Can be 'live', 'dead', 'wind', or 'snow'. The default is 'live'.
        name : string, optional
            Name of the load. The default is 'Load'.
        f : float, optional
            scaling factor for the load. The default is 1.0.

        Returns
        -------
        None.

        """
        
        
        self.load_id = load_id
        self.name = name
        self.ltype = ltype
        self.f = f
    
    def factored(self,comb_type='ULS',lead=True):
        
        factor = 1.0
        
        # variable loads
        if self.ltype == 'live' or self.ltype == 'wind' or self.ltype == 'snow':
            if comb_type == 'ULS':
                factor = 1.5
            
            if not lead:
                # load is not leading, so it must be multiplied by
                # corresponding factor. These factors apply also in
                # characteristic combation of the serviceability limit state
                if self.ltype == 'live':
                    factor *= 0.7
                elif self.ltype == 'snow':
                    factor *= 0.7
                elif self.ltype == 'wind':
                    factor *= 0.6
        # permanent loads
        elif self.ltype == 'dead':
            if comb_type == 'ULS':
                factor = 1.15
        
        self.f = factor
            

class PointLoad(Load):
    """ Class for point loads (forces and moments) """

    def __init__(self, node, v, load_id=LoadIDs['ULS'], ltype='live',
                 name='PointLoad',f=1.0):
        """        

        Parameters
        ----------
        node : FrameNode object
            Node of point of load application.
        v : numpy array or list
            Vector of the load.
        load_id : TYPE, optional
            Identifier for the load to be used in structural analysis. The default is LoadIDs.ULS.
        ltype : string, optional
            Load type. The default is 'live'.
        name : string, optional
            Name of the load. The default is 'PointLoad'.
        f : float, optional
            Scaling factor. The default is 1.0.

        Returns
        -------
        None.

        """
        
        super().__init__(load_id, ltype, name, f)
    
        
        self.node = node
        self.v = np.asarray(v)        

    @property
    def coordinate(self):
        return self.node.coords

    def add_load(self, fem_model):
        """ Adds load to a FEM model """        
        fem_model.add_load(fem.PointLoad(self.load_id, self.node.fem_node, self.v, self.f))


class LineLoad(Load):
    def __init__(self, member, values, direction, load_id=LoadIDs['ULS'], f=1.0,
                 ltype='live', name='LineLoad',coord_sys="global"):
        """ Class for line loads. Line load can be uniform or trapezoidal
        
            input:
                member .. FrameMember type object that is the subject of the load
                values .. list type object. values[0] and values[1] are the magnitudes
                          of the load at the ends of the member
                direction .. 'x' or 'y'
                load_id .. integer type load identifier
                f .. scaling factor
                ltype .. type of load
                name .. name of load
                coord_sys .. "global" or 'local'. If 'local' the load is given in the
                             local coordinate system of the member.
        """
        super().__init__(load_id, ltype, name, f)
        
        self.member = member
        self.member.loads.append(self)
        
        #self.member.has_load = True
        #self.member.q0 = values[0]
        #self.member.q1 = values[1]            
        
        
        self.values = np.asarray(values)
        self.direction = direction
        self.f = f
        self.element_ids = None
        self.coord_sys = coord_sys

    @property
    def coordinates(self):
        return self.member.coords()

    def calc_k(self):
        """ Evaluate the change in the value of a line load
            per element on a member
        """
        v0, v1 = self.values
        k = (v1 - v0) / self.member.num_elements
        return k

    def add_load(self, fem_model):
        """ Add the load to the FEM model """
        #k = self.calc_k()
        v0, v1 = self.values
        #qvals = self.f*np.linspace(v0,v1,self.member.num_elements+1)        
        qvals = self.f*np.linspace(v0,v1,len(self.member.fem_elements)+1)        
        for i, elem_id in enumerate(self.member.fem_elements):
            #v1 = (i * k) + v0
            load = fem.LineLoad(self.load_id,
                                elem_id,
                                [0.0, 1.0],
                                [qvals[i],qvals[i+1]],                                
                                self.direction,
                                self.coord_sys)
            fem_model.add_load(load)
            #v0 = v1

class PWLineLoad(Load):
    """ Class for piecewise constant line loads. There can be two distint zones """
    def __init__(self, member, values, xdiv, direction, load_id=LoadIDs['ULS'], f=1.0,
                 ltype='live', name='LineLoad',coord_sys="global"):
        """ Class for line loads. Line load can be uniform or trapezoidal
        
            input:
                member .. FrameMember type object that is the subject of the load
                values .. list type object. values[0] and values[1] are the magnitudes
                          of the load at the intervals
                xdiv .. point of step. This is given as a local coordinate of the member
                direction .. 'x' or 'y'
                load_id .. integer type load identifier
                f .. scaling factor
                ltype .. type of load
                name .. name of load
                coord_sys .. "global" or 'local'. If 'local' the load is given in the
                             local coordinate system of the member.
        """
        super().__init__(load_id, ltype, name, f)
        
        self.member = member
        self.member.loads.append(self)
        
        #self.member.has_load = True
        #self.member.q0 = values[0]
        #self.member.q1 = values[1]            
        
        
        self.values = np.asarray(values)
        self.xdiv = xdiv
        self.direction = direction
        self.f = f
        self.element_ids = None
        self.coord_sys = coord_sys

    @property
    def coordinates(self):
        return self.member.coords()

    def calc_k(self):
        """ Evaluate the change in the value of a line load
            per element on a member
        """
        v0, v1 = self.values
        k = (v1 - v0) / self.member.num_elements
        return k

    def add_load(self, fem_model):
        """ Add the load to the FEM model """
        #k = self.calc_k()
        qval = self.f*self.values
        xdiv = self.member.local_to_global(self.xdiv)[0]
        
        for elem in self.member.fem_elements:
            X = [elem.nodes[0].x,elem.nodes[1].x]
            X.sort()
            if X[1] <= xdiv:
                q = qval[0]
            elif X[0] >= xdiv:
                q = qval[1]
            elif X[0] < xdiv and X[1] >= xdiv:
                nd = np.argmax(abs(qval))
                q = qval[nd]
                
            load = fem.LineLoad(self.load_id,elem,
                                [0.0, 1.0],
                                [q,q],                                
                                self.direction,
                                self.coord_sys)
            fem_model.add_load(load)
            #v0 = v1



from matplotlib import pyplot as plt

import metku.frame2d as f2d
import metku.truss2d as t2d
from metku.optimization.optimzation_enums import ObjectiveEnum
from metku.optimization.structopt import *
from metku.structures.steel.steel_member import SteelMember


class TrussProblem(OptimizationProblem):
    """
    Class for truss optimization problem
    """

    def __init__(self,
                 name='TrussProblem',
                 displacement_limit: float = 30,
                 variable_groups=None,
                 objective: callable or ObjectiveEnum = ObjectiveEnum.WEIGHT,
                 structure: f2d.Frame2D = None,
                 include_joint_strength: bool = False,
                 include_joint_geometry: bool = False,
                 include_section_strength: bool = True,
                 include_stability: bool = True,
                 include_section_class: bool = True,
                 include_displacement: bool = True,
                 load_id_SLS=LoadIDs.SLS_Charasteristic,
                 load_id_ULS=LoadIDs.ULS,
                 penalty_fun: callable = None):

        super().__init__(name, structure=structure)
        self.variable_groups = variable_groups
        self.objective = objective
        self.include_joint_strength = include_joint_strength
        self.include_joint_geometry = include_joint_geometry
        self.include_section_strength = include_section_strength
        self.include_stability = include_stability
        self.include_section_class = include_section_class
        self.include_displacement = include_displacement
        self.displacement_limit = displacement_limit
        self.load_id_SLS = load_id_SLS
        self.load_id_ULS = load_id_ULS
        # Penalty function
        if penalty_fun is None:
            penalty_fun = lambda *x: 0
        self.penalty_fun = penalty_fun
        # Create variables
        self.create_variables()
        # Create constraints
        self.create_constraints()

        # Assign objective
        match self.objective:
            case ObjectiveEnum.WEIGHT:
                self.obj = self.weight_fun
            case ObjectiveEnum.COST:
                raise NotImplementedError("Cost Objective onot implemented")
            case callable(self.objective):
                self.obj = self.objective
            case _:
                raise ValueError("Objective must be either "
                                 "ObjectiveTypeEnum or callable")

    @property
    def weight_fun(self):
        """
        Function for calculating structure's weight
        :return: ObjectiveFunction
        """
        if self.structure is not None:
            def obj_fun(x):
                return self.structure.weight + self.penalty_fun(x)

            obj = ObjectiveFunction(name="Objective ",
                                    obj_fun=obj_fun)
            obj.problem = self

            return obj

        else:
            raise ValueError("Structure is not defined!")

    def create_variables(self):
        """
        Creates design variables
        """
        self.vars = []
        for vg in self.variable_groups:
            for var in vg.variables:
                self.add(var)

    def joint_geometry_constraints(self, joint):
        """
        Creates joint geometry constraint functions
        :param joint: TrussJoint
        :return: constraint functions in a dict
        """

        con_funs = {}

        if joint.joint_type == 'Y':
            w1 = list(joint.webs.values())[0]

            def b(x):
                """
                Web's breadth constraint
                """
                return w1.cross_section.B / joint.chord.cross_section.B - 1

            def b_min(x):
                """
                Minimum breadth for web 1
                """
                val = 0.35 * joint.chord.cross_section.B
                val /= w1.cross_section.B
                val -= 1
                return val

            def theta_min(x):

                val = 30 / np.degrees(joint.theta1) - 1
                # val = 30 - np.degrees(w1.angle - joint.chord.angle)
                return val

            con_funs = {
                f'Joint {joint.jid} b': b,
                f'Joint {joint.jid} b1_min': b_min,
                f'Joint {joint.jid} theta1_min': theta_min,
            }

        elif joint.joint_type == 'K':
            w1, w2 = joint.webs.values()

            def beta(x):
                """
                Beta constraint
                """
                b1 = w1.cross_section.B
                b2 = w2.cross_section.B
                h1 = w1.cross_section.H
                h2 = w2.cross_section.H
                b0 = joint.chord.cross_section.B

                return (b1 + b2 + h1 + h2) / (4 * b0) - 1
                # return (b1 + b2 + h1 + h2) - (4 * b0)

            def b1b0(x):
                b1 = w1.cross_section.B
                b0 = joint.chord.cross_section.B
                t0 = joint.chord.cross_section.T

                return (0.1 + 0.01 * b0 / t0) / (b1 / b0) - 1

            def b2b0(x):
                b2 = w2.cross_section.B
                b0 = joint.chord.cross_section.B
                t0 = joint.chord.cross_section.T

                return (0.1 + 0.01 * b0 / t0) / (b2 / b0) - 1

            def g_min(x):
                """
                Minimum value for gap
                """
                gap = w1.cross_section.T + w2.cross_section.T
                val = gap - joint.g1
                val /= abs(joint.g1)
                return val

            def g_min_beta(x):
                """
                Minimum value for gap
                """
                beta = joint.rhs_joint.beta()
                val = 0.5 * (1 - beta) * joint.chord.cross_section.B
                val /= abs(joint.g1)
                val -= 1
                return val

            def g_max_beta(x):
                """
                Maximum value for gap
                """
                beta = joint.rhs_joint.beta()
                val = joint.g1
                val /= (1.5 * (1 - beta) * joint.chord.cross_section.B)
                val -= 1
                return val

            def b1_min(x):
                """
                Minimum breadth for web 1
                """
                val = 0.35 * joint.chord.cross_section.B
                val /= w1.cross_section.B
                val -= 1
                return val

            def b1_max(x):
                val = -0.85 * joint.chord.cross_section.B
                val /= w1.cross_section.B
                val += 1
                return val

            def b2_min(x):
                val = 0.35 * joint.chord.cross_section.B
                val /= w2.cross_section.B
                val -= 1
                return val

            def b2_max(x):
                val = -0.85 * joint.chord.cross_section.B
                val /= w2.cross_section.B
                val += 1
                return val

            def theta1_min(x):

                val = 30 / np.degrees(joint.theta1) - 1
                # val = 30 - np.degrees(w1.angle - joint.chord.angle)
                return val

            def theta1_max(x):
                val = np.degrees(joint.theta1) / 90 - 1
                # val = 30 - np.degrees(w1.angle - joint.chord.angle)

                return val

            def theta2_min(x):
                val = 30 / np.degrees(joint.theta2) - 1
                # val = 30 - np.degrees(w2.angle - joint.chord.angle)
                return val

            def theta2_max(x):
                val = np.degrees(joint.theta2) / 90 - 1
                # val = 30 - np.degrees(w2.angle - joint.chord.angle)
                return val

            def e_pos(x):
                return joint.e / ((0.25 * joint.chord.cross_section.H)) - 1

                # return joint.e - ((0.25 * joint.chord.cross_section.H))

            def e_neg(x):
                return -1 - joint.e / (0.55 * joint.chord.cross_section.H)

                # return  - joint.e - (0.55 * joint.chord.cross_section.H)

            con_funs = {
                f'Joint {joint.jid} beta': beta,
                f'Joint {joint.jid} b1b0': b1b0,
                f'Joint {joint.jid} b2b0': b2b0,
                f'Joint {joint.jid} g_min': g_min,
                f'Joint {joint.jid} g_min_beta': g_min_beta,
                f'Joint {joint.jid} g_max_beta': g_max_beta,
                f'Joint {joint.jid} b1_min': b1_min,
                f'Joint {joint.jid} b1_max': b1_max,
                f'Joint {joint.jid} b2_min': b2_min,
                f'Joint {joint.jid} b2_max': b2_max,
                f'Joint {joint.jid} theta1_min': theta1_min,
                f'Joint {joint.jid} theta1_max': theta1_max,
                f'Joint {joint.jid} theta2_min': theta2_min,
                f'Joint {joint.jid} theta2_max': theta2_max,
                f'Joint {joint.jid} e_pos': e_pos,
                f'Joint {joint.jid} e_neg': e_neg}

        elif joint.joint_type == 'KT':
            # TODO
            pass

        return con_funs

    def joint_strength_constraints(self, joint):
        """
        Creates joint strength costraint functions

        :param joint: TrussJoint to be calculated
        :return: constraint functions in a dict
        """
        con_funs = {}

        if joint.joint_type == 'Y':
            w1 = list(joint.webs.values())[0]

            def chord_face_failure(x):
                NRd1 = joint.rhs_joint.chord_face_failure()
                return w1.NEd[self.load_id_ULS] / NRd1 - 1

            def chord_web_buckling(x):
                NRd1 = joint.rhs_joint.chord_web_buckling()
                return -w1.NEd[self.load_id_ULS] / NRd1 - 1

            def brace_failure(x):
                N1Rd = joint.rhs_joint.brace_failure()
                return abs(w1.NEd[self.load_id_ULS]) / N1Rd - 1

            con_funs = {
                f'Joint {joint.jid} chord face failure': chord_face_failure,
                f'Joint {joint.jid} chord web buckling': chord_web_buckling,
                f'Joint {joint.jid} brace failure': brace_failure,
            }


        elif joint.joint_type == 'K':
            w1, w2 = joint.webs.values()

            def chord_face_failure_1(x):

                NRd1, NRd2 = joint.rhs_joint.chord_face_failure()
                return abs(w1.NEd[self.load_id_ULS]) / NRd1 - 1
                # return abs(w1.ned) - NRd1

            def chord_face_failure_2(x):
                NRd1, NRd2 = joint.rhs_joint.chord_face_failure()
                return abs(w2.NEd[self.load_id_ULS]) / NRd2 - 1
                # return abs(w2.ned) - NRd2

            def punching_shear_1(x):
                NRd1 = joint.rhs_joint.punching_shear()[0]
                return abs(w1.NEd[self.load_id_ULS]) / NRd1 - 1
                # return abs(w1.ned) - NRd1

            def punching_shear_2(x):
                NRd2 = joint.rhs_joint.punching_shear()[1]
                return abs(w2.NEd[self.load_id_ULS]) / NRd2 - 1
                # return abs(w2.ned) - NRd2

            def chord_shear_0(x):
                joint.rhs_joint.V0 = joint.V0[self.load_id_ULS]
                (N1Rd, N2Rd), N0Rd = joint.rhs_joint.chord_shear()

                return joint.N0[self.load_id_ULS] / N0Rd - 1
                # return joint.N0 - N0Rd

            def chord_shear_1(x):
                joint.rhs_joint.V0 = joint.V0[self.load_id_ULS]
                (N1Rd, N2Rd), N0Rd = joint.rhs_joint.chord_shear()

                return abs(w1.NEd[self.load_id_ULS]) / N1Rd - 1

            def chord_shear_2(x):
                joint.rhs_joint.V0 = joint.V0[self.load_id_ULS]
                (N1Rd, N2Rd), N0Rd = joint.rhs_joint.chord_shear()

                return abs(w2.NEd[self.load_id_ULS]) / N2Rd - 1

            def brace_failure_1(x):
                N1Rd = joint.rhs_joint.brace_failure()[0]
                return abs(w1.NEd[self.load_id_ULS]) / N1Rd - 1
                # return abs(w1.ned) - N1Rd

            def brace_failure_2(x):
                N2Rd = joint.rhs_joint.brace_failure()[1]
                return abs(w2.NEd[self.load_id_ULS]) / N2Rd - 1
                # return abs(w2.ned) - N2Rd

            con_funs = {
                f'Joint {joint.jid} chord face failure 1': chord_face_failure_1,
                f'Joint {joint.jid} chord face failure 2': chord_face_failure_2,
                f'Joint {joint.jid} punching shear 1': punching_shear_1,
                f'Joint {joint.jid} punching shear 2': punching_shear_2,
                f'Joint {joint.jid} chord shear 0': chord_shear_0,
                # f'Joint {joint.jid} chord shear 1': chord_shear_1,
                # f'Joint {joint.jid} chord shear 2': chord_shear_2,
                f'Joint {joint.jid} brace failure 1': brace_failure_1,
                f'Joint {joint.jid} brace failure 2': brace_failure_2,
            }

        elif joint.joint_type == 'KT':
            # TODO
            pass

        return con_funs

    def cross_section_constraints(self, mem, elem, load_id=LoadIDs.ULS, idx=0):
        """
        Creates cross-section constraints
        :return:
        """

        def absmax(vals):
            min_val = min(vals)
            max_val = max(vals)
            abs_max = max(abs(min_val), abs(max_val))
            if abs_max == abs(min_val):
                return min_val
            else:
                return max_val

        def compression(x):
            NEd = elem.fint[load_id]['fx'][idx]
            return -NEd / mem.NRd - 1

        def tension(x):
            NEd = elem.fint[load_id]['fx'][idx]
            return NEd / mem.NRd - 1

        def shear(x):
            VEd = elem.fint[load_id]['fz'][idx]
            return abs(VEd) / mem.VRd - 1

        def bending_moment(x):
            # Moment about y
            # TODO: Moment about z
            MEd = elem.fint[load_id]['my'][idx]
            return abs(MEd) / mem.cross_section.plastic_bending_resistance() - 1

        def shear_moment(x):

            gammaM0 = 1
            VEd = abs(elem.fint[load_id]['fz'][idx])
            VRd = mem.cross_section.VRd
            if VEd <= 0.5 * VRd:
                rho = 0
            else:
                rho = (2 * VEd / VRd - 1) ** 2
            MRd = mem.MRd
            t = mem.cross_section.T
            Wply = mem.cross_section.Wply
            Av = mem.cross_section.Ashear[0]
            MVRd = min((Wply - rho * Av ** 2 / (8 * t) / gammaM0) * mem.fy, MRd)
            MEd = abs(elem.fint[load_id]['my'][idx])
            return MEd / MVRd - 1

        # VARMALLA PUOLELLA OLEVA OLETUS PL 3 & 4
        # def NVM(x):
        #     """
        #     Shear + Normal + Moment
        #     """
        #     NEd = abs(elem.fint[load_id]['fx'][idx])
        #     VEd = abs(elem.fint[load_id]['fz'][idx])
        #     MEd = abs(elem.fint[load_id]['my'][idx])
        #     NRd = mem.cross_section.NRd
        #     VRd = mem.cross_section.VRd
        #     MRd = mem.cross_section.MRd
        #     rho = (2 * VEd / VRd - 1) ** 2
        #
        #     return NEd / NRd + MEd / MRd - (1 - rho)

        # POIKKILEIKKAUSLUOKILLE 1 & 2
        def NVM(x):
            """
            Shear + Normal + Moment
            """
            gammaM0 = 1
            NEd = abs(elem.fint[load_id]['fx'][idx])
            VEd = abs(elem.fint[load_id]['fz'][idx])
            MEd = abs(elem.fint[load_id]['my'][idx])
            VRd = mem.cross_section.VRd
            MRd = mem.cross_section.MRd
            A = mem.cross_section.A
            t = mem.cross_section.T
            b = mem.cross_section.B
            if VEd <= 0.5 * VRd:
                rho = 0
            else:
                rho = (2 * VEd / VRd - 1) ** 2
            Wply = mem.cross_section.Wply
            Av = mem.cross_section.Ashear[0]
            Awred = (1 - rho) * (A - 2 * b * t)
            Atotred = A - rho * (A - 2 * b * t)
            NVRd = Atotred * mem.fy / gammaM0
            MVRd = min((Wply - rho * Av ** 2 / (8 * t) / gammaM0) * mem.fy,
                       MRd)
            nV = NEd / NVRd
            aV = min(Awred / Atotred, 0.5)

            if NEd <= 0.25 * NVRd and NEd <= (0.5 * Awred * mem.fy / gammaM0):
                MNVRd = MVRd
            else:
                MNVRd = MVRd * (1 - nV) / (1 - 0.5 * aV)

            return MEd / MNVRd - 1

        # Cross-Section constraints
        return {
            'Ncd': compression,
            'Ntd': tension,
            'VEd': shear,
            'MEd Y': bending_moment,
            'VEd + MEd': shear_moment,
            'NEd + VEd + MEd': NVM
        }

    def stability_constraints(self, mem: f2d.FrameMember, smem: SteelMember,
                              load_id=LoadIDs.ULS) -> dict[str: callable]:
        """

        :param mem: FrameMember -object
        :param smem: SteelMember -object
        :param section_class:
        :param load_id:
        :return:
        """
        def buckling_y(x):
            mem.update_steel_members_forces(load_id)
            smem.check_buckling()
            return smem.check_buckling()[0] - 1

        def buckling_z(x):
            mem.update_steel_members_forces(load_id)
            smem.check_buckling()
            return smem.check_buckling()[1] - 1

        def com_compression_bending_y(x):
            mem.update_steel_members_forces(load_id)
            section_class = mem.cross_section.section_class()
            return smem.check_beamcolumn(section_class=section_class)[0] - 1

        def com_compression_bending_z(x):
            mem.update_steel_members_forces(load_id)
            section_class = mem.cross_section.section_class()
            return smem.check_beamcolumn(section_class=section_class)[1] - 1

        # ADD ANY MISSING CONSTRAINTS HERE
        def my_own_stability_con_fun(x):
            ...

        return {
            'Buckling Y': buckling_y,
            'Buckling Z': buckling_z,
            'Stab. NEd + MEd Y': com_compression_bending_y,
            'Stab. NEd + MEd Z': com_compression_bending_z,
            # 'My stability test': my_own_stability_con_fun
        }

    def deflection_constraints(self, truss: t2d.Truss2D):
        """
        Creates deflection constraint functions
        :param mem:
        :return:
        """

        def disp_fun(x):
            truss.calculate(load_id=self.load_id_SLS)
            max_deflection = truss.get_deflection(LoadIDs.SLS_Charasteristic)
            # To make sure correct values are used in stability design
            truss.calculate(load_id=self.load_id_ULS)
            return max_deflection - self.displacement_limit

        return disp_fun

    def cross_section_class_constraints(self, mem):
        """
        Cross-section class constraints
        :param mem:
        :return:
        """
        # SECTION IN PURE COMPRESSION
        def section_class(x):
            b_ = mem.cross_section.H - 3 * mem.cross_section.T
            c_t = b_ / mem.cross_section.T
            return c_t / (38 * mem.cross_section.epsilon) - 1

        return section_class

    def create_constraints(self):
        """
        Cretes constraints and saves them to self.cons
        """
        # Initialize cons as an empty list
        self.cons = []
        # MEMBER AND CROSS-SECTION CONSTRAINTS
        for mem in self.structure.members.values():
            # STABILITY CONSTRAINTS
            if self.include_stability:
                for j, smem in enumerate(mem.members):
                    stability_con_dict = self.stability_constraints(mem, smem)
                    for stab_con_name, stab_con_fun in stability_con_dict.items():
                        if "MEd" in stab_con_name and "chord" in mem.mtype:
                            self.non_linear_constraint(name=f"{stab_con_name} {mem.mtype} {mem.mem_id}|{j}",
                                                       con_fun=stab_con_fun,
                                                       fea_required=True)
                        elif "Buckling" in stab_con_name:
                            self.non_linear_constraint(name=f"{stab_con_name} {mem.mtype} {mem.mem_id}|{j}",
                                                       con_fun=stab_con_fun,
                                                       fea_required=True)
            # CROSS-SECTION CLASS CONSTRAINT
            if self.include_section_class:
                sect_fun = self.cross_section_class_constraints(mem)
                sect_con = self.non_linear_constraint(sect_fun,
                                                      name=f"Section Class {mem.mtype} {mem.mem_id} ")
            # CROSS-SECTION STRENGTH
            if self.include_section_strength:
                for i, (key, elem) in enumerate(mem.elements.items()):
                    con_fun_dict = self.cross_section_constraints(mem, elem, load_id=LoadIDs.ULS, idx=0)

                    for con_name, con_fun in con_fun_dict.items():
                        self.non_linear_constraint(con_fun, name=f"{con_name}: {mem.mtype} {mem.mem_id}|{key}", fea_required=True)
                    # Last element's end node
                    if i == len(mem.elements) - 1 and mem.mtype != 'web':
                        con_fun_dict = self.cross_section_constraints(mem, elem, load_id=LoadIDs.ULS, idx=1)

                        for con_name, con_fun in con_fun_dict.items():
                            self.non_linear_constraint(con_fun, name=f"{con_name}: {mem.mtype} {key}|{i + 1}")
        # DISPLACEMENT
        if self.include_displacement:
            disp_fun = self.deflection_constraints(self.structure)
            disp_con = self.non_linear_constraint(
                con_fun=disp_fun, name=" Truss Deflection")
            self.add(disp_con)
        # JOINT CONS
        if self.include_joint_geometry or self.include_joint_strength:
            for joint in self.structure.joints.values():
                # Geometry Constraints
                if self.include_joint_geometry:
                    con_funs = self.joint_geometry_constraints(joint)
                    for name, con_fun in con_funs.items():
                        self.non_linear_constraint(con_fun=con_fun,
                                                   name=name)
                # Strength Constraints
                if self.include_joint_strength:
                    strength_con_funs = self.joint_strength_constraints(joint)
                    for name, con_fun in strength_con_funs.items():
                        self.non_linear_constraint(con_fun=con_fun,
                                                   name=name)


def create_truss(h1: float,
                 h2: float,
                 L: float,
                 n: int,
                 dx: float,
                 qd: float = 0,
                 qk: float = 0) -> t2d.Truss2D:
    simple = dict(
        H0=0,
        H1=h2,
        H2=h1,
        L1=L / 2,
        n=n,
        dx=dx
    )
    truss = t2d.Truss2D(simple=simple)
    for mem in truss.members.values():
        if mem.mtype == 'top_chord':
            truss.add(f2d.LineLoad(mem, [qd, qd], 'y', load_id=LoadIDs.ULS))
            truss.add(f2d.LineLoad(mem, [qk, qk], 'y', load_id=LoadIDs.SLS_Charasteristic))
    truss.add(f2d.XYHingedSupport([0, truss.H1]))
    truss.add(f2d.YHingedSupport([truss.L, truss.H1]))
    truss.generate()
    truss.calculate('ALL')
    return truss


if __name__ == '__main__':
    from metku.optimization.variable_group import VariableGroup, VariableTypeEnum
    from metku.sections.steel.catalogue import shs_profiles
    from metku.optimization.solvers import SLP

    # THESE ARE NOT SORTED VALUES
    SHS_PROFILE_NAMES = list(shs_profiles.keys())

    L = 24000
    truss = create_truss(h1=L / 10,
                         h2=L / 10 - (1 / 20 * L / 2),
                         L=L,
                         dx=L/20,
                         n=16,
                         qd=-21.5,
                         qk=-15)

    top_chords = truss.top_chords
    bottom_chords = truss.bottom_chords
    webs = list(truss.webs.values())

    tc_sections = [mem.cross_section for mem in top_chords]
    bc_sections = [mem.cross_section for mem in bottom_chords]
    web_sections = [mem.cross_section for mem in webs]

    tc_group = VariableGroup(name="Top Chord",
                             var_type=VariableTypeEnum.INDEX,
                             attribute="profile",
                             objects=top_chords,
                             values=SHS_PROFILE_NAMES)

    bc_group = VariableGroup(name="Bottom Chord",
                             var_type=VariableTypeEnum.INDEX,
                             attribute="profile",
                             objects=bottom_chords,
                             values=SHS_PROFILE_NAMES)

    web_group = VariableGroup(name="Webs",
                              var_type=VariableTypeEnum.INDEX,
                              attribute="profile",
                              objects=webs,
                              values=SHS_PROFILE_NAMES)

    tc_HBT = VariableGroup(name="Top Chords",
                           var_type=VariableTypeEnum.CONTINUOUS,
                           attributes=[["H", "B"], "T"],
                           objects=tc_sections,
                           lower_bounds=[100, 3],
                           upper_bounds=[300, 15],
                           value=[160, 8])
    bc_HBT = VariableGroup(name="Bottom Chords",
                           var_type=VariableTypeEnum.CONTINUOUS,
                           attributes=[["H", "B"], "T"],
                           objects=bc_sections,
                           lower_bounds=[100, 3],
                           upper_bounds=[300, 15],
                           value=[160, 6])
    web_HBT = VariableGroup(name="Webs",
                            var_type=VariableTypeEnum.CONTINUOUS,
                            attributes=[["H", "B"], "T"],
                            objects=web_sections,
                            lower_bounds=[50, 3],
                            upper_bounds=[200, 15],
                            value=[100, 5])

    groups = [tc_HBT, bc_HBT]
    # Number of webs
    nw = int(len(webs) / 2)
    for i in range(nw):
        w1 = webs[2 * i]
        w2 = webs[2 * i + 1]
        WEB_group = VariableGroup(name="Webs",
                                  var_type=VariableTypeEnum.CONTINUOUS,
                                  attributes=[["H", "B"], "T"],
                                  objects=[w1.cross_section, w2.cross_section],
                                  lower_bounds=[40, 3],
                                  upper_bounds=[200, 8],
                                  value=[100, 5])
        groups.append(WEB_group)

    problem = TrussProblem(name="Truss Optimization",
                           structure=truss,
                           displacement_limit=L/250,  # mm
                           include_joint_strength=False,
                           include_joint_geometry=False,
                           include_stability=True,
                           include_displacement=True,
                           include_section_class=True,
                           include_section_strength=True,
                           variable_groups=groups)

    x0 = [var.value for var in problem.vars]
    problem(x0, ncons=20)
    solver = SLP([0.05, 0.05], beta=100, update_beta=False)
    fopt, xopt = solver.solve(problem, x0=x0, maxiter=250, verb=True)
    problem(xopt)
    truss.plot_normal_force(load_id=LoadIDs.ULS)


# Author(s): Kristo Mela
# This source code is licensed under the MIT license. See LICENSE in the repository root directory.
# Copyright 2022 Kristo Mela
# -*- coding: utf-8 -*-
"""
Created on Mon Jan  4 12:48:16 2021

Design rules of EN 1995-1-1, Design of Timber Structures

@author: Viktor Haimi
"""
try:
    from metku.eurocodes.en1995.constants import *
except:
    from eurocodes.en1995.constants import *

kmod_data = {'solid_timber': {1: {'perm': 0.6, 'lt': 0.7, "mt": 0.8, "st": 0.9, "inst": 1.1},
                              2: {'perm': 0.6, 'lt': 0.7, "mt": 0.8, "st": 0.9, "inst": 1.1},
                              3: {'perm': 0.5, 'lt': 0.55, "mt": 0.65, "st": 0.7, "inst": 0.9}
                              },
             'glt': {1: {'perm': 0.6, 'lt': 0.7, "mt": 0.8, "st": 0.9, "inst": 1.1},
                     2: {'perm': 0.6, 'lt': 0.7, "mt": 0.8, "st": 0.9, "inst": 1.1},
                     3: {'perm': 0.5, 'lt': 0.55, "mt": 0.65, "st": 0.7, "inst": 0.9}
                     },
             'lvl': {1: {'perm': 0.6, 'lt': 0.7, "mt": 0.8, "st": 0.9, "inst": 1.1},
                     2: {'perm': 0.6, 'lt': 0.7, "mt": 0.8, "st": 0.9, "inst": 1.1},
                     3: {'perm': 0.5, 'lt': 0.55, "mt": 0.65, "st": 0.7, "inst": 0.9}
                    }
             }

kdef_data = {'solid_timber': {1: 0.6, 2: 0.8, 3: 2.0},
             'glt': {1: 0.6, 2: 0.8, 3: 2.0},
             'lvl': {1: 0.6, 2: 0.8, 3: 2.0}
             }


def kmod(material: str, service_class: int, load_duration_class: str) -> float:
    """
    SFS-EN 1995-1-1 § 3, taulukko 3.1
    @param material:                valittu materiaali
    @param service_class:           käyttöluokka
    @param load_duration_class:     aikaluokka
    @return:                        palauttaa k_mod arvon
    """
    return kmod_data[material][service_class][load_duration_class]


def kdef(material: str, service_class: int) -> float:
    """
    SFS-EN 1995-1-1 § 3, taulukko 3.2
    @param material:        valittu materiaali
    @param service_class:   käyttöluokka
    @return:                palauttaa k_def arvon
    """
    return kdef_data[material][service_class]


def k_to_d(f_k: float, kmod: float, safety_factor: float) -> float:
    return (kmod * f_k) / safety_factor


def get_gammaM(timber_type: str) -> float:
    if timber_type == 'solid_timber':
        return gammaM_st
    elif timber_type == 'glt':
        return gammaM_glt
    elif timber_type == 'lvl':
        return gammaM_lvl
    else:
        return gammaM_other



# -*- coding: utf-8 -*-
# Copyright 2022 Kristo Mela
# This source code is licensed under the MIT license. See LICENSE in the repository root directory.
# Author(s): Kristo Mela
"""
Created on Thu Mar 15 18:14:54 2018

Euler-Bernoulli beam element in 2D

@author: kmela
"""

import numpy as np
from functools import lru_cache

CACHE_BOUND = 2**10

#try:
#    from metku.framefem import Element, LineLoad, FEMNode
#except:
#    from framefem import Element, LineLoad

from ..framefem import Element, LineLoad

class EBBeam(Element):
    """ Euler-Bernoulli beam element

        Parameters:
        -----------
        :param n1: start node
        :param n2: end node
        :param section: element's section properties
        :param material: element's material properties

        :type n1: FEMNode
        :type n2: FEMNode
        :type section: Section
        :type material: Material

        Variables:
        ----------
        :ivar bending_moment: bending moment values at element's ends
        :ivar shear_force: shear force values at element's ends

        :vartype bending_moment: list
        :vartype shear_force: list
    """

    def __init__(self, n1, n2, section, material):
        Element.__init__(self, n1, n2, section, material)

        #self.bending_moment = [0.0, 0.0]
        #self.shear_force = [0.0, 0.0]
        
        self.Krel = {'K11':None, 'K12':None, 'K21':None, 'K22':None}


    def transformation_matrix(self,initial=True):
        """ Calculates transformation matrix from local to global coordinates
            Chandrupatla 3rd. Edition, p. 249
            Returns:
            --------
            :return: element's transformation matrix
            :rtype: np.array

         """
        c = self.direction_cosines(initial)
        L = np.zeros((6, 6))
        L[0, 0:2] = c[0:2]
        L[1, 0] = -L[0, 1]
        L[1, 1] = L[0, 0]
        L[2, 2] = 1.0
        L[3:6, 3:6] = L[0:3, 0:3]
        return L


    def init_dofs(self):
        """ For beam members all degrees of freedom
            are in play.
        """
        for n in self.nodes:
            n.dofs[:3] = 0
                        

    def global_dofs(self):
        """ Get numbering of element's global degrees of freedom

            Returns:
            --------
            :return: list of end nodes dofs added together
            :rtype: list
        """
        return np.append(self.nodes[0].dofs,self.nodes[1].dofs)
        #return self.nodes[0].dofs + self.nodes[1].dofs

    @lru_cache(CACHE_BOUND)
    def local_stiffness_matrix(self, E, A, I1, Le):
        """ Generates stiffness matrix in local coordinates


            Returns:
            --------
            :return: element's stiffness matrix in local coordinates
            :rtype: np.array

        """
        rodc = E * A / Le
        EI1 = E * I1
        k0 = np.zeros((6, 6))
        #
        k0[[0, 3], [0, 3]] = rodc
        k0[[0, 3], [3, 0]] = -rodc
        #
        bend1 = 12 * EI1 / Le ** 3
        k0[[1, 4], [1, 4]] = bend1
        k0[[1, 4], [4, 1]] = -k0[1, 1]
        #
        bend2 = 6 * EI1 / Le ** 2
        k0[[1, 2, 1, 5], [2, 1, 5, 1]] = bend2
        #
        k0[[2, 4, 4, 5], [4, 2, 5, 4]] = -bend2
        #
        k0[[2, 5], [5, 2]] = 2 * EI1 / Le
        #
        k0[[2, 5], [2, 5]] = 4 * EI1 / Le
        
        if len(self.releases) > 0:
            """ There are releases in the element, so the stiffness matrix
                is modified.
            """
            rel = self.releases
            # nrel is the list of non-released forces
            nrel = np.setdiff1d(np.arange(6),rel)
            
            K22 = k0[np.ix_(rel,rel)]
            K11 = k0[np.ix_(nrel,nrel)]
            K12 = k0[np.ix_(nrel,rel)]
            K21 = k0[np.ix_(rel,nrel)]
            
            self.Krel['K11'] = K11
            self.Krel['K12'] = K12
            self.Krel['K21'] = K21
            self.Krel['K22'] = K22
            
            """
            print(rel,nrel)
            print('k0')
            print(k0)
            print('K22')
            print(K22)
            print('K11')
            print(K11)
            print('K12')
            print(K12)
            print('K21')
            print(K21)
            """
            
            #K221K21inv = np.linalg.inv(K22).dot(K21)
            
            kc = K11 - K12.dot(np.linalg.inv(K22).dot(K21))
            
            """
            print('kc = ')
            print(kc)
            print(kc.shape)
            """
            k0[np.ix_(nrel,nrel)] = kc
            k0[rel,:] = 0
            k0[:,rel] = 0
            #print(k0)
            

        return k0

    def stiffness_matrix(self,initial=True):
        """ Computes element's global stiffness matrix

            Returns:
            ---------
            :return: element's stiffness matrix in global coordinates
            :rtype: np.array

        """
        E = self.material.young
        A = self.section.A
        I1 = self.section.I[0]
        Le = self.length(initial)
        k0 = self.local_stiffness_matrix(E, A, I1, Le)

        # k0 = CheckReleases[fem.elem[N],k0];

        # local-global matrix
        L = self.transformation_matrix(initial)

        # globaali elementin matriisi
        ke = L.transpose().dot(k0.dot(L))

        return ke

    def local_geometric_stiffness_matrix(self,lcase=0,initial=True) -> np.ndarray:
        """ Geometric stiffness matrix in local coordinates
            source: Cook et. al 1989, Section 14.2

            Returns:
            ---------
            :return: element's geometric stiffness matrix in local coordinates
            :rtype: np.ndarray

        """
        #P = self.axial_force[1]
        P = self.fint[lcase]['fx'][1]
        Le = self.length(initial)

        g0 = P / 30 / Le
        keg0 = g0 * np.array([[36, 3 * Le, -36, 3 * Le],
                              [3 * Le, 4 * Le ** 2, -3 * Le, -Le ** 2],
                              [-36, -3 * Le, 36, -3 * Le],
                              [3 * Le, -Le ** 2, -3 * Le, 4 * Le ** 2]])
        kg0 = np.zeros([6, 6])
        q = [1, 2, 4, 5]
        kg0[np.ix_(q, q)] = keg0

        return kg0

    def geometric_stiffness_matrix(self,lcase=0,initial=True):
        """ Geometric stiffness matrix in global coordinates \n
            From: Cook et. al (1989), Section 14.2

            Returns:
            ---------
            :return: element's geometric stiffness matrix in global coordinates
            :rtype: np.array

        """

        # local-global transformation matrix
        L = self.transformation_matrix(initial)

        kG0 = self.local_geometric_stiffness_matrix(lcase,initial)

        kG = L.transpose().dot(kG0.dot(L))

        return kG
    
    def local_second_order_stiffness_matrix(self,lcase=0):
        """ Stiffness matrix for second order analysis
            From: Salmi & Kuula (2012), Rakenteiden mekaniikka, Section 8.3
        """
        E = self.material.young
        A = self.section.A
        I1 = self.section.I[0]
        Le = self.length()
        
        P = self.fint[lcase]['fx'][1]
        
        if P < 0:
            # For compression members, use second order theory
            P = abs(P)
        
            rodc = E * A / Le
            a = 4*E*I1/Le
            b = 2*E*I1/Le
            c = 6*E*I1/Le**2
            d = 12*E*I1/Le**3
                    
            k = np.sqrt(P/E/I1)
            lam = k*Le
            
            slam = np.sin(lam)
            clam = np.cos(lam)
            
            D = 2-2*clam-lam*slam
            K = lam**3/12*slam/D
            H = lam**2/6*(1-clam)/D
            PSI = lam/4*(slam-lam*clam)/D
            PHI = 0.5*lam*(lam-slam)/D
            
            k0 = np.zeros((6, 6))
            #
            k0[[0, 3], [0, 3]] = rodc
            k0[[0, 3], [3, 0]] = -rodc
            #
            #bend1 = 12 * E*I1 / Le ** 3
            k0[[1, 4], [1, 4]] = d*K
            k0[[1, 4], [4, 1]] = -k0[1, 1]
            #
            #bend2 = 6 * E*I1 / Le ** 2
            k0[[1, 2, 1, 5], [2, 1, 5, 1]] = c*H
            #
            k0[[2, 4, 4, 5], [4, 2, 5, 4]] = -c*H
            #
            k0[[2, 5], [5, 2]] = b*PHI
            #
            k0[[2, 5], [2, 5]] = a*PSI
        else:
            # For members in tension, use first order stiffness matrix
            k0 = self.local_stiffness_matrix(E, A, I1, Le)
        
        if len(self.releases) > 0:
            """ There are releases in the element, so the stiffness matrix
                is modified.
            """
            rel = self.releases
            # nrel is the list of non-released forces
            nrel = np.setdiff1d(np.arange(6),rel)
            
            K22 = k0[np.ix_(rel,rel)]
            K11 = k0[np.ix_(nrel,nrel)]
            K12 = k0[np.ix_(nrel,rel)]
            K21 = k0[np.ix_(rel,nrel)]
            
            self.Krel['K11'] = K11
            self.Krel['K12'] = K12
            self.Krel['K21'] = K21
            self.Krel['K22'] = K22
            
            """
            print(rel,nrel)
            print('k0')
            print(k0)
            print('K22')
            print(K22)
            print('K11')
            print(K11)
            print('K12')
            print(K12)
            print('K21')
            print(K21)
            """
            
            #K221K21inv = np.linalg.inv(K22).dot(K21)
            
            kc = K11 - K12.dot(np.linalg.inv(K22).dot(K21))
            
            """
            print('kc = ')
            print(kc)
            print(kc.shape)
            """
            k0[np.ix_(nrel,nrel)] = kc
            k0[rel,:] = 0
            k0[:,rel] = 0
            #print(k0)
            

        return k0
    
    def second_order_stiffness_matrix(self,lcase=0):
        # local-global transformation matrix
        L = self.transformation_matrix()

        k0 = self.local_second_order_stiffness_matrix(lcase)

        k = L.transpose().dot(k0.dot(L))

        return k


    def equivalent_nodal_loads(self, load, order=1, initial=True):
        """ Equivalent nodal loads for a load 
        
            'load' is a Load type object. By default, it is assumed that
            'load' is LineLoad

            Returns:
            ---------
            :return: Equivalent nodal loads for load in vector q
            :rtype: np.array
        """
        floc = np.zeros(6)

        T = self.transformation_matrix(initial)
        L = self.length(initial)
        
        #print(isinstance(load,LineLoad))
        
        if isinstance(load,LineLoad):            
            q1 = load.f*load.qval[0]
            q2 = load.f*load.qval[1]
            #print(q1,q2)
            if load.coords == 'local':
                #print('local coordinates')
                """ Load is given in local coordinates """
                if load.dir == "x":
                    """ Load is in the axial direction """
                elif load.dir == 'y':
                    """ Load is perpendicular to the member """
                    if order == 1:
                        floc[1] = 7/20*q1*L + 3/20*q2*L
                        floc[4] = 3/20*q1*L + 7/20*q2*L                    
                        floc[2] = L**2*(1/20*q1+1/30*q2)
                        floc[5] = -L**2*(1/30*q1+1/20*q2)
                    else:
                        # For second order analysis, assume uniform
                        # load on element
                        P = self.fint[load.sid]['fx'][0]
                        if P >= 0:                        
                            floc[1] = 7/20*q1*L + 3/20*q2*L
                            floc[4] = 3/20*q1*L + 7/20*q2*L                    
                            floc[2] = L**2*(1/20*q1+1/30*q2)
                            floc[5] = -L**2*(1/30*q1+1/20*q2)
                        else:
                            k = np.sqrt(abs(P)/self.section.E/self.section.Iy)
                            a = 0.5*L
                            lam = k*a
                            s = np.sin(k*L)
                            c = np.cos(k*L)
                            
                            floc[1] = 0.5*q1*L
                            floc[4] = 0.5*q1*L                    
                            floc[2] = q1*a**2*(lam*c-s)/lam**2/s
                            floc[5] = -q1*a**2*(lam*c-s)/lam**2/s
                #print(floc)
            else:
                """ Load is given in global coordinates: it has to be transformed
                    first into local coordinates.
                """
                if load.dir == "x":
                    #q = np.array([q1, 0, 0])
                    q = np.array([1, 0, 0])
                else:
                    #q = np.array([0, q1, 0])
                    q = np.array([0, 1, 0])
                # Load vector transformed into element local coordinate system
                qloc = load.f*T[:3, :3].dot(q)
                                
                if order == 1:
                    """ Loads perpendicular to the axis of the element """
                    qy1 = q1*qloc[1]
                    qy2 = q2*qloc[1]
                    floc[1] = 7/20*qy1*L + 3/20*qy2*L
                    floc[4] = 3/20*qy1*L + 7/20*qy2*L
                    floc[2] = L**2*(1/20*qy1+1/30*qy2)
                    floc[5] = -L**2*(1/30*qy1+1/20*qy2)
                
                    """ Loads parallel to the axis of the element """
                    qx1 = q1*qloc[0]
                    qx2 = q2*qloc[0]
                    floc[0] = 0.5*L*(0.5*(qx1+qx2)-1/6*(qx2-qx1))
                    floc[3] = 0.5*L*(0.5*(qx1+qx2)+1/6*(qx2-qx1))
                else:
                    # For second order analysis, assume uniform
                    # load on element
                    P = self.fint[load.sid]['fx'][0]
                    if P >= 0:                        
                        """ Loads perpendicular to the axis of the element """
                        qy1 = q1*qloc[1]
                        qy2 = q2*qloc[1]
                        floc[1] = 7/20*qy1*L + 3/20*qy2*L
                        floc[4] = 3/20*qy1*L + 7/20*qy2*L
                        floc[2] = L**2*(1/20*qy1+1/30*qy2)
                        floc[5] = -L**2*(1/30*qy1+1/20*qy2)
                    
                        """ Loads parallel to the axis of the element """
                        qx1 = q1*qloc[0]
                        qx2 = q2*qloc[0]
                        floc[0] = 0.5*L*(0.5*(qx1+qx2)-1/6*(qx2-qx1))
                        floc[3] = 0.5*L*(0.5*(qx1+qx2)+1/6*(qx2-qx1))
                    else:
                        k = np.sqrt(abs(P)/self.section.E/self.section.Iy)
                        a = 0.5*L
                        lam = k*a
                        s = np.sin(lam)
                        c = np.cos(lam)
                        
                        qy1 = q1*qloc[1]
                        qy2 = q2*qloc[1]
                        print(f'lambda = {lam:4.3f}')
                        print(f's = {s:4.3f}')
                        print(f'c = {c:4.3f}')
                        #print(k,lam,s,c,(lam*c-s)/lam**2/s)
                        #print()
                        floc[1] = 0.5*qy1*L
                        floc[4] = 0.5*qy1*L                    
                        floc[2] = -qy1*a**2*(lam*c-s)/lam**2/s
                        floc[5] = qy1*a**2*(lam*c-s)/lam**2/s
                        
                        
                        
                        """ Loads parallel to the axis of the element """
                        qx1 = q1*qloc[0]
                        qx2 = q2*qloc[0]
                        floc[0] = 0.5*L*(0.5*(qx1+qx2)-1/6*(qx2-qx1))
                        floc[3] = 0.5*L*(0.5*(qx1+qx2)+1/6*(qx2-qx1))
                # Construct nodal load in local coordinates
                # qloc[0,1,2] = axial, shear, moment of node 1
                # qloc[3,4,5] = axial, shear, moment ofnode 2
                # Axial force
                #floc[[0, 3]] = 0.5 * qloc[0] * L
                # Shear force
                #floc[[1, 4]] = 0.5 * qloc[1] * L
                # Moment
                #floc[2] = qloc[1] * L ** 2 / 12.0
                #floc[5] = -floc[2]

                #print(floc)

        if len(self.releases) > 0:
            """ There are releases in the element, so the local load vector
                is modified.
            """
            rel = self.releases
            # nrel is the list of non-released forces
            nrel = np.setdiff1d(np.arange(6),rel)            
            
            #print(np.linalg.inv(self.Krel['K22']).dot(floc[rel]))
            floc[nrel] -= self.Krel['K12'].dot(np.linalg.inv(self.Krel['K22']).dot(floc[rel]))
            floc[rel] = 0            

        # Store local loads
        self.floc[load.sid] = floc

        # print(self.floc)

        fglob = T.transpose().dot(floc)
        
        
        return fglob

    def shape_fun(self, x):
        """ Evaluate Hermitean shape functions at local coordinate x

            Returns:
            ---------
            :return: Hermitean shape functions at local coordinate x
            :rtype: np.array


        """
        L = self.length()
        N = np.zeros(4)
        N[0] = 1 - 3 * x ** 2 / L ** 2 + 2 * x ** 3 / L ** 3
        N[1] = x - 2 * x ** 2 / L + x ** 3 / L ** 2
        N[2] = 3 * x ** 2 / L ** 2 - 2 * x ** 3 / L ** 3
        N[3] = -x ** 2 / L ** 2 + x ** 3 / L ** 2

        return N

    def internal_forces(self,lcase=0,initial=True):
        """ Calculate internal forces
            NOTE: these internal forces do not take
            loads along the element into account!

            Works only for a single load case!
        """

        """ Get nodal displacements in local coordinates
            and multiply them with local element stiffness matrix
            to get internal forces in member's local coordinate system.
        """
        q = self.local_displacements(lcase)
        # ke = self.k0
        E = self.material.young
        A = self.section.A
        I1 = self.section.I[0]
        Le = self.length(initial)
        ke = self.local_stiffness_matrix(E, A, I1, Le)

        try:            
            R = ke.dot(q) - self.floc[lcase]
        except:
            R = ke.dot(q)

        """ Any load on the element not acting on a node must be
            taken into account here. This requires the following steps:
            1. Identify loads acting on element
            2. Compute their equivalent nodal loads
            3. Modify internal forces

            Probably an easy solution is to include an
            Attribute floc for equivalent nodal loads. The nodal loads are
            saved there, when they are first calculated for analysis
        """

        """
        self.axial_force[:2] = [-R[0], R[3]]
        self.bending_moment[:2] = [-R[2], R[5]]
        self.shear_force[:2] = [R[1], -R[4]]
        """
        
        self.fint[lcase]['fx'] = [-R[0], R[3]]
        self.fint[lcase]['my'] = [-R[2], R[5]]
        self.fint[lcase]['fz'] = [R[1], -R[4]]
        
    def second_order_internal_forces(self,lcase=0):
        """ Calculate internal forces in second order analysis
            NOTE: these internal forces do not take
            loads along the element into account!

            Works only for a single load case!
        """

        """ Get nodal displacements in local coordinates
            and multiply them with local element stiffness matrix
            to get internal forces in member's local coordinate system.
        """
        q = self.local_displacements(lcase,order=2)
        ke = self.local_second_order_stiffness_matrix(lcase)
        
        try:            
            R = ke.dot(q) - self.floc[lcase]
        except:
            R = ke.dot(q)

        """ Any load on the element not acting on a node must be
            taken into account here. This requires the following steps:
            1. Identify loads acting on element
            2. Compute their equivalent nodal loads
            3. Modify internal forces

            Probably an easy solution is to include an
            Attribute floc for equivalent nodal loads. The nodal loads are
            saved there, when they are first calculated for analysis
        """

        """
        self.axial_force[:2] = [-R[0], R[3]]
        self.bending_moment[:2] = [-R[2], R[5]]
        self.shear_force[:2] = [R[1], -R[4]]
        """
        
        self.fint2[lcase]['fx'] = [-R[0], R[3]]
        self.fint2[lcase]['my'] = [-R[2], R[5]]
        self.fint2[lcase]['fz'] = [R[1], -R[4]]
        
class EBBeam3D(Element):
    """ Euler-Bernoulli beam element in 3D

        Parameters:
        -----------
        :param n1: start node
        :param n2: end node
        :param nref: reference node for orienting the element
        :param section: element's section properties
        :param material: element's material properties

        :type n1: FEMNode
        :type n2: FEMNode
        :type section: Section
        :type material: Material

        Variables:
        ----------
        :ivar bending_moment: bending moment values at element's ends
        :ivar shear_force: shear force values at element's ends

        :vartype bending_moment: list
        :vartype shear_force: list
        
        Degrees of freedom
        [0] .. ux1
        [1] .. uy1
        [2] .. uz1
        [3] .. rx1
        [4] .. ry1
        [5] .. rz1
        
        [6] .. ux2
        [7] .. uy2
        [8] .. uz2
        [9] .. rx2
        [10] .. ry2
        [11] .. rz2
        
        TODO:
            equivalent_nodal_loads
            internal_forces
            test all functions
    """
    
    def __init__(self, n1, n2, section, material, nref = None):
        
        
        Element.__init__(self, n1, n2, section, material)
        
        self.dim = 3
        
        """ Reference node for orienting the member
            This node must not be on the line passing through nodes n1 and n2.
        """
        if nref is None:
            """ Calculate third point separately 
                The 2D case is in xz-plane, so by default, the y
                axis is the out-of-plane axis. The member is oriented
                such that in the plane case, the local and global y axes
                coincide.
            """
            if np.linalg.norm(np.cross(self.direction_vector(),[0,1,0])) < 1e-6:
            #if abs(n2.coord[1]-n1.coord[1]) < 1e-6:
                """ local x axis coincides with the global
                    y axis, so the reference point is obtained by stepping
                    in the z direction
                """
                nref = FEMNode(-1,n2.coord[0],n2.coord[1],n2.coord[2]+1)
            elif np.linalg.norm(np.cross(self.direction_vector(),[0,0,1])) < 1e-6:
                """ member is vertial, so local x axis coincides with
                    z-axis                
                """
                nref = FEMNode(-1,n2.coord[0]+1,n2.coord[1],n2.coord[2])
            else:
                nref = FEMNode(-1,n2.coord[0],n2.coord[1],n2.coord[2]+1)
                
            
        self.nref = nref
        
    def init_dofs(self):
        """ For beam members all degrees of freedom
            are in play.
            
        """
        for n in self.nodes:
            n.dofs[:6] = 0
    
    def direction_vector(self):
        
        return super().direction_cosines()
            
    def direction_cosines(self):
        """ Calculates direction cosines for a 3D-beam element """
        
        l0 = np.zeros((3,3))
        
        X = self.coord()
        L = self.length()
        dX = X[1, :] - X[0, :]        
        # direction cosines along the local x axis
        l0[0,:] = dX / L
        """
        l = l0[0,0]
        m = l0[0,1]
        n = l0[0,2]
        D = np.sqrt(l**2+n**2)
        l0[2,:] = 1/D*np.array([n,0,-l])
        l0[1,:] = 1/D*np.array([-m*n,n*l+n**2,-m*n])
        """
        # calculate direction cosines along the local y axis
        # coordinates of the reference node
        X3 = self.nref.coord
        dX13 = X3-X[0,:]
        l13 = np.linalg.norm(dX13)
        
        # unit vector between point 1 and reference point
        V13 = dX13/l13
        VZ0 = np.cross(V13,l0[0,:])        
        l0[1,:] = VZ0/np.linalg.norm(VZ0)
        
        l0[2,:] = np.cross(l0[0,:],l0[1,:])
        
        return l0
        
    def transformation_matrix(self):
        """ Calculates transformation matrix from local to global coordinates

            Returns:
            --------
            :return: element's transformation matrix
            :rtype: np.array

         """
        l0 = self.direction_cosines()
        L = np.zeros((12, 12))
        L[0:3, 0:3] = l0
        L[3:6, 3:6] = l0
        L[6:9, 6:9] = l0
        L[9:12, 9:12] = l0
        return L

    def global_dofs(self):
        """ Get numbering of element's global degrees of freedom

            Returns:
            --------
            :return: list of end nodes dofs added together
            :rtype: list
        """
        return np.append(self.nodes[0].dofs,self.nodes[1].dofs)
        #return self.nodes[0].dofs + self.nodes[1].dofs
    
    
    def local_stiffness_matrix(self):
        """ Generates stiffness matrix in local coordinates


            Returns:
            --------
            :return: element's stiffness matrix in local coordinates
            :rtype: np.array

        """
        E = self.material.young
        G = self.material.shear_modulus
        A = self.section.A
        I1 = self.section.Iy
        I2 = self.section.Iz
        J = self.section.It
        
        Le = self.length()

        rodc = E * A / Le
        EI1 = E * I1
        EI2 = E * I2
        torc = G*J/Le
        k0 = np.zeros((12, 12))
        # Axial stiffness
        k0[[0, 6], [0, 6]] = rodc
        k0[[0, 6], [6, 0]] = -rodc
        # Torsional stiffness
        k0[[3, 9], [3, 9]] = torc
        k0[[3, 9], [9, 3]] = -torc
        # Bending with respect to y axis
        bend1y = 12 * EI2 / Le ** 3
        k0[[1, 7], [1, 7]] = bend1y
        k0[[1, 7], [7, 1]] = -bend1y
        #
        bend2y = 6 * EI2 / Le ** 2
        k0[[1, 1, 5, 11], [5, 11, 1, 1]] = bend2y
        #
        k0[[5, 7, 7, 11], [7, 5, 11, 7]] = -bend2y
        #
        k0[[5, 11], [11, 5]] = 2 * EI2 / Le
        #
        k0[[5, 11], [5, 11]] = 4 * EI2 / Le
        
        bend1z = 12 * EI1 / Le**3
        k0[[2, 8], [2, 8]] = bend1z
        k0[[8, 2], [2, 8]] = -bend1z

        bend2z = 6 * EI1 / Le ** 2
        k0[[8, 4, 8, 10], [4, 8, 10, 8]] = bend2z
        #
        k0[[2, 10, 2, 4], [10, 2, 4, 2]] = -bend2z

        k0[[4, 10], [10, 4]] = 2 * EI1 / Le
        #
        k0[[4, 10], [4, 10]] = 4 * EI1 / Le

        return k0
    
    def stiffness_matrix(self):
        """ Computes element's global stiffness matrix

            Returns:
            ---------
            :return: element's stiffness matrix in global coordinates
            :rtype: np.array

        """
        k0 = self.local_stiffness_matrix()

        # k0 = CheckReleases[fem.elem[N],k0];

        # local-global matrix
        L = self.transformation_matrix()

        # globaali elementin matriisi
        ke = L.transpose().dot(k0.dot(L))

        return ke
    
    def equivalent_nodal_loads(self, load):
        """ Equivalent nodal loads for a load 
        
            'load' is a Load type object. By default, it is assumed that
            'load' is LineLoad

            Returns:
            ---------
            :return: Equivalent nodal loads for load in vector q
            :rtype: np.array
        """

        """ Equivalent nodal loads in local coordinates.
            Six degrees of freedom per node.
            
            Degrees of freedom
            [0] .. ux1
            [1] .. uy1
            [2] .. uz1
            [3] .. rx1
            [4] .. ry1
            [5] .. rz1
            
            [6] .. ux2
            [7] .. uy2
            [8] .. uz2
            [9] .. rx2
            [10] .. ry2
            [11] .. rz2
        """
        floc = np.zeros(12)

        T = self.transformation_matrix()
        L = self.length()

        if isinstance(load,LineLoad):
            q1 = load.qval[0]
            q2 = load.qval[1]
            #print(q1,q2)
            if load.coords == 'local':
                #print('local coordinates')
                """ Load is given in local coordinates """
                if load.dir == "x":
                    """ Load is in the axial direction """
                elif load.dir == 'y':
                    """ Load is perpendicular to the member """
                    floc[1] = 7/20*q1*L + 3/20*q2*L
                    floc[7] = 3/20*q1*L + 7/20*q2*L                    
                    floc[5] = L**2*(1/20*q1+1/30*q2)
                    floc[11] = -L**2*(1/30*q1+1/20*q2)
                elif load.dir == 'z':
                    floc[2] = 7/20*q1*L + 3/20*q2*L
                    floc[8] = 3/20*q1*L + 7/20*q2*L                    
                    floc[4] = L**2*(1/20*q1+1/30*q2)
                    floc[10] = -L**2*(1/30*q1+1/20*q2)
                #print(floc)
            else:
                """ Load is given in global coordinates: it has to be transformed
                    first into local coordinates.
                """
                if load.dir == "x":                    
                    q = np.array([1, 0, 0,0,0,0])
                elif load.dir == 'y':                    
                    q = np.array([0, 1, 0,0,0,0])
                elif load.dir == 'z':                    
                    q = np.array([0, 0, 1,0,0,0])
                # Load vector transformed into element local coordinate system
                qloc = T[:3, :3].dot(q)
                
                """ Loads perpendicular to the axis of the element """
                qy1 = q1*qloc[1]
                qy2 = q2*qloc[1]
                floc[1] = 7/20*qy1*L + 3/20*qy2*L
                floc[4] = 3/20*qy1*L + 7/20*qy2*L
                floc[2] = L**2*(1/20*qy1+1/30*qy2)
                floc[5] = -L**2*(1/30*qy1+1/20*qy2)
                
                """ Loads parallel to the axis of the element """
                qx1 = q1*qloc[0]
                qx2 = q2*qloc[0]
                floc[0] = 0.5*L*(0.5*(qx1+qx2)-1/6*(qx2-qx1))
                floc[3] = 0.5*L*(0.5*(qx1+qx2)+1/6*(qx2-qx1))
                
                # Construct nodal load in local coordinates
                # qloc[0,1,2] = axial, shear, moment of node 1
                # qloc[3,4,5] = axial, shear, moment ofnode 2
                # Axial force
                #floc[[0, 3]] = 0.5 * qloc[0] * L
                # Shear force
                #floc[[1, 4]] = 0.5 * qloc[1] * L
                # Moment
                #floc[2] = qloc[1] * L ** 2 / 12.0
                #floc[5] = -floc[2]

                #print(floc)

        if len(self.releases) > 0:
            """ There are releases in the element, so the local load vector
                is modified.
            """
            rel = self.releases
            # nrel is the list of non-released forces
            nrel = np.setdiff1d(np.arange(12,rel))          
            
            #print(np.linalg.inv(self.Krel['K22']).dot(floc[rel]))
            floc[nrel] -= self.Krel['K12'].dot(np.linalg.inv(self.Krel['K22']).dot(floc[rel]))
            floc[rel] = 0            

        # Store local loads
        self.floc[load.sid] = floc

        # print(self.floc)

        fglob = T.transpose().dot(floc)
        
        
        return fglob

    def local_geometric_stiffness_matrix(self,lcase=0):
        """ Geometric stiffness matrix in local coordinates
            source: Cook et. al 1989, Section 14.2

            Returns:
            ---------
            :return: element's geometric stiffness matrix in local coordinates
            :rtype: np.array

        """
        #P = self.axial_force[1]
        P = self.fint[lcase]['fx'][1]
        Le = self.length()

        g0 = P / 30 / Le
        keg0 = g0 * np.array([[36, 3 * Le, -36, 3 * Le],
                              [3 * Le, 4 * Le ** 2, -3 * Le, -Le ** 2],
                              [-36, -3 * Le, 36, -3 * Le],
                              [3 * Le, -Le ** 2, -3 * Le, 4 * Le ** 2]])
        kg0 = np.zeros([6, 6])
        q = [1, 2, 4, 5]
        kg0[np.ix_(q, q)] = keg0

        return kg0

    def internal_forces(self,lcase=0):
            """ Calculate internal forces
    
            """
    
            """ Get nodal displacements in local coordinates
                and multiply them with local element stiffness matrix
                to get internal forces in member's local coordinate system.
            """
            q = self.local_displacements(lcase)
            ke = self.local_stiffness_matrix()
    
            #print('q=')
            #print(q)
    
            try:            
                R = ke.dot(q) - self.floc[lcase]
            except:
                R = ke.dot(q)
    
            """ Any load on the element not acting on a node must be
                taken into account here. This requires the following steps:
                1. Identify loads acting on element
                2. Compute their equivalent nodal loads
                3. Modify internal forces
    
                Probably an easy solution is to include an
                Attribute floc for equivalent nodal loads. The nodal loads are
                saved there, when they are first calculated for analysis
            """
    
            """
            self.axial_force[:2] = [-R[0], R[3]]
            self.bending_moment[:2] = [-R[2], R[5]]
            self.shear_force[:2] = [R[1], -R[4]]
            Degrees of freedom
                [0] .. ux1
                [1] .. uy1
                [2] .. uz1
                [3] .. rx1
                [4] .. ry1
                [5] .. rz1
                
                [6] .. ux2
                [7] .. uy2
                [8] .. uz2
                [9] .. rx2
                [10] .. ry2
                [11] .. rz2
            """
            self.fint[lcase]['fx'] = [-R[0], R[6]]
            self.fint[lcase]['fy'] = [-R[1], R[7]]
            self.fint[lcase]['fz'] = [R[2], -R[8]]
            self.fint[lcase]['mx'] = [-R[3], R[9]]
            self.fint[lcase]['my'] = [-R[4], R[10]]
            self.fint[lcase]['mz'] = [R[5], -R[11]]


# Author(s): Kristo Mela
# This source code is licensed under the MIT license. See LICENSE in the repository root directory.
# Copyright 2022 Kristo Mela
# -*- coding: utf-8 -*-
"""
Created on Fri Jun 12 09:46 2020

EN 1993-1-2 Supplementary rules for fire design

@author: kmela
"""

import math
import numpy as np

from metku.eurocodes.en1993.constants import gammaM0, gammaM1, gammaM2
from metku.materials import steel_data as sd
    
# EN 1993-1-2: Table 3.1
REDUCTION_FACTORS_CS = {20:{'ky':1.0,'kp':1.0,'kE':1.0,'kp02':1.0},
                        100:{'ky':1.0,'kp':1.0,'kE':1.0,'kp02':1.0},
                        200:{'ky':1.0,'kp':0.807,'kE':0.9,'kp02':0.89},
                        300:{'ky':1.0,'kp':0.613,'kE':0.8,'kp02':0.78},
                        400:{'ky':1.0,'kp':0.420,'kE':0.7,'kp02':0.65},
                        500:{'ky':0.780,'kp':0.360,'kE':0.6,'kp02':0.53},
                        600:{'ky':0.470,'kp':0.180,'kE':0.310,'kp02':0.30},
                        700:{'ky':0.230,'kp':0.075,'kE':0.130,'kp02':0.13},
                        800:{'ky':0.110,'kp':0.050,'kE':0.090,'kp02':0.07},
                        900:{'ky':0.060,'kp':0.0375,'kE':0.0675,'kp02':0.05},
                        1000:{'ky':0.040,'kp':0.0250,'kE':0.0450,'kp02':0.03},
                        1100:{'ky':0.020,'kp':0.0125,'kE':0.0225,'kp02':0.02},
                        1200:{'ky':0.000,'kp':0.0,'kE':0.0,'kp02':0.0},
                        }

# EN 1993-1-2, Table C.1
REDUCTION_FACTORS_14301 = {20:{'ky':0.26,'k02':1.0,'ku':1.0,'kE':1.0,'kEct':0.11,'epsU': 0.40},
                           100:{'ky':0.24,'k02':0.82,'ku':0.87,'kE':0.96,'kEct':0.05,'epsU': 0.40},
                           200:{'ky':0.19,'k02':0.68,'ku':0.77,'kE':0.92,'kEct':0.02,'epsU': 0.40},
                           300:{'ky':0.19,'k02':0.64,'ku':0.73,'kE':0.88,'kEct':0.02,'epsU': 0.40},
                           400:{'ky':0.19,'k02':0.60,'ku':0.72,'kE':0.84,'kEct':0.02,'epsU': 0.40},
                           500:{'ky':0.19,'k02':0.54,'ku':0.67,'kE':0.80,'kEct':0.02,'epsU': 0.40},
                           600:{'ky':0.22,'k02':0.49,'ku':0.58,'kE':0.76,'kEct':0.02,'epsU': 0.35},
                           700:{'ky':0.26,'k02':0.40,'ku':0.43,'kE':0.71,'kEct':0.02,'epsU': 0.30},
                           800:{'ky':0.35,'k02':0.27,'ku':0.27,'kE':0.63,'kEct':0.02,'epsU': 0.20},
                           900:{'ky':0.38,'k02':0.14,'ku':0.15,'kE':0.45,'kEct':0.02,'epsU': 0.20},
                           1000:{'ky':0.040,'k02':0.06,'ku':0.07,'kE':0.2,'kEct':0.02,'epsU': 0.20},
                           1100:{'ky':0.040,'k02':0.03,'ku':0.03,'kE':0.1,'kEct':0.02,'epsU': 0.20},
                           1200:{'ky':0.040,'k02':0.0,'ku':0.00,'kE':0.0,'kEct':0.02,'epsU': 0.20},
                           }

# EN 1993-1-2, Table D.1
REDUCTION_FACTORS_BOLTS = {20:{'kb':1.0},
                           100:{'kb':0.968},
                           150:{'kb':0.952},
                           200:{'kb':0.935},
                           300:{'kb':0.903},
                           400:{'kb':0.775},
                           500:{'kb':0.550},
                           600:{'kb':0.220},
                           700:{'kb':0.100},
                           800:{'kb':0.067},
                           900:{'kb':0.033},
                           1000:{'kb':0.0},
                           }

SPECIAL_TEMPS = np.array([20,100,150,200,300,400,500,600,700,800,900,1000,1100,1200])




def standard_fire(t):
    """ Returns the temperature of standard fire ISO 834 
        input:
            t: time in minutes
        output:
            Temperature in Celsius
    """
    return 20 + 345*math.log(8*t+1,10)

def hydrocarbon_fire(t):
    """ Returns the temperature of hydrocarbon fire
        input:
            t: time in minutes
        output:
            Temperature in Celsius
    """
    return 20 + 1080*(1-0.325*math.exp(-0.167*t)-0.675*math.exp(-2.5*t))

def linear_interpolation(p1,p2,x):
    """ Performs linear interpolation.
        p1 = (x1,y1), p2 = (x2,y2)
        
        The function finds the value y corresponding to x1 <= x <= x2:
        Such that y(x) = k*x + b
    """
    
    y1 = p1[1]
    x1 = p1[0]
    y2 = p2[1]
    x2 = p2[0]
    
    k = (y2-y1)/(x2-x1)
    b = y1-k*x1
    
    return k*x+b

def reduce_property(t,prop,steel='carbon'):
    """ Function for determining the reduction factor at temperature 't'
        of property 'prop'
    """
    rf = 1.0
    # Check if 't' matches any of the special temperatures
    if any(SPECIAL_TEMPS == t):
        if steel == 'carbon':
            rf = REDUCTION_FACTORS_CS[t][prop]
        elif steel =='1.4301':
            rf = REDUCTION_FACTORS_14301[t][prop]
        elif steel == 'bolts':
            rf = REDUCTION_FACTORS_BOLTS[t][prop]
        else:
            print("Unidentified steel grade!")
    else:
        # t does not match any of the special temperatures;
        # Do linear interpolation
        
        """ Find the largest smaller and smallest larger
            temperatures and the corresponding reduction factors
        """
        t_smaller = np.max(SPECIAL_TEMPS[SPECIAL_TEMPS < t])
        t_larger = np.min(SPECIAL_TEMPS[SPECIAL_TEMPS > t])
        
        if steel == 'carbon':
            p_smaller = REDUCTION_FACTORS_CS[t_smaller][prop]
            p_larger = REDUCTION_FACTORS_CS[t_larger][prop]
        elif steel == '1.4301':
            p_smaller = REDUCTION_FACTORS_14301[t_smaller][prop]
            p_larger = REDUCTION_FACTORS_14301[t_larger][prop]
        elif steel == 'bolts':
            p_smaller = REDUCTION_FACTORS_BOLTS[t_smaller][prop]
            p_larger = REDUCTION_FACTORS_BOLTS[t_larger][prop]
            
        rf = linear_interpolation((t_smaller,p_smaller), (t_larger,p_larger), t)
    
    return rf

""" Chapter 3: Materials """
def EaT(t,Ea=210000,steel='carbon'):
    """ Temperature-dependent Young's modulus
        for carbon steel or stainless steel
        input:
            t .. temperature [Celsius]
    """
    kE = reduce_property(t,'kE',steel)
    return kE*Ea

def fyT(t,fy=355,steel='carbon'):
    """ Temperature-dependent yield strength
        for carbon steel
        input:
            t .. temperature [Celsius]
    """
    ky = reduce_property(t,'ky',steel)
    return ky*fy

def fpT(t,fp=355,steel='carbon'):
    """ Temperature-dependent yield strength
        for carbon steel
        input:
            t .. temperature [Celsius]
    """
    kp = reduce_property(t,'kp',steel)
    return kp*fp

def f02pT(t,fy=355,steel="1.4301"):
    """ Temperature-dependent value for proof strength at 0.2 plastic strain """

    k = reduce_property(t,'k02',steel)
    
    return k*fy

def fuT(t,fy_T,steel='carbon'):
    """ EN 1993-1-2, Annex A(2) 
        Ultimate strength of carbon steel, when taking
        into account work hardening
    """
    
    if steel=='carbon':    
        if t < 300:
            fu_T = 1.25*fy_T
        elif t < 400:
            fu_T = fy_T*(2-0.0025*t)
        else:
            fu_T = fy_T
    else:
        ku = reduce_property(t,'ku',steel)
        fu_T = ku*fy_T
        
    return fu_T

def kbT(t):
    """ Temperature-dependent value for bolt shear and tension resistance """

    k = reduce_property(t,'kb','bolts')    
    return k
    

def stress(strain,t,Ea,fya,hardening=False):
    """ Returns the stress of carbon steel for given
        strain and
        temperature 't'
        
        EN 1993-1-2, Fig. 3.1
    """
    eps_yt = 0.02
    eps_tt = 0.15
    eps_ut = 0.20
    
    # Proportional limit and Young's modulus at temperature 't'
    fp_t = fpT(t,fya)
    Ea_t = EaT(t,Ea)
    
    eps_pt = fp_t/Ea_t
    
    if strain <= eps_pt:
        st = Ea_t*strain
    elif strain < eps_yt:        
        fy_t = fyT(t,fya)
        c = (fy_t-fp_t)**2/((eps_yt-eps_pt)*Ea_t-2*(fy_t-fp_t))
        b = math.sqrt(c*(eps_yt-eps_pt)*Ea_t+c**2)
        a = math.sqrt((eps_yt-eps_pt)*(eps_yt-eps_pt+c/Ea_t))
        st = fp_t - c + (b/a)*math.sqrt(a**2-(eps_yt-strain)**2)
    elif strain <= eps_tt:
        fy_t = fyT(t,fya)
        if hardening:
            fu_T = fuT(t,fy_t)
            if strain < 0.04:
                st = 50*(fu_T-fy_t)*strain + 2*fy_t - fu_T
            else:
                st = fu_T
        else:
            st = fy_t            
    elif strain < eps_ut:
        fy_t = fyT(t,fya)
        if hardening:
            fu_T = fuT(t,fy_t)
            st = fu_T*(1-20*(strain-0.15))
        else:
            st = fy_t*(1-(strain-eps_tt)/(eps_ut-eps_tt))
    else:
        st = 0.0
        
    return st

def stress_stainless(strain,t,Ea,fya,fua,steel="1.4301"):
    """ Stress-strain relationship of stainless steels at elevated temperatures """
    
    # Proportional limit and Young's modulus at temperature 't'
    f02p_t = f02pT(t,fya,steel)
    
    Ea_t = EaT(t,Ea,steel)
    
    
    eps_ct = f02p_t/Ea_t + 0.002
    kE_ct = reduce_property(t,'kEct',steel)
    Ect_t = kE_ct*Ea
    
    eps_ut = reduce_property(t,'epsU',steel)
     
    
    if strain <= eps_ct:        
        b = (1-eps_ct*Ect_t/f02p_t)*Ea_t*eps_ct/(Ea_t*eps_ct/f02p_t-1)/f02p_t
        a = (Ea_t*eps_ct - f02p_t)/f02p_t/pow(eps_ct,b)
        st = Ea_t*strain/(1+a*pow(strain,b))
    else:
        fu_t = fuT(t,fua,steel)
        e = (fu_t - f02p_t)**2/((eps_ut-eps_ct)*Ect_t-2*(fu_t-f02p_t))
        d = math.sqrt(e*(eps_ut-eps_ct)*Ect_t + e**2)
        c = math.sqrt((eps_ut-eps_ct)*(eps_ut-eps_ct + e/Ect_t))
        st = f02p_t - e + d/c*math.sqrt(c**2 - (eps_ut-strain)**2)
    
    
    return st
    

def therm_elongation(t):
    """ Relative thermal elongation of steel
        EN 1993-1-2, 3.4.1.1
    """
    
    if t < 750:
        dL = 1.2e-5*t + 0.4e-8*t**2-2.416e-4
    elif t <= 860:
        dL = 1.1e-2
    else:
        dL = 2e-5*t -6.2e-3
    
    return dL

def therm_conductivity(t):
    """ Thermal conductivity of steel [W/mK]
        EN 1993-1-2, 3.4.1.3
    """
    if t < 800:
        lambda_a = 54-3.33e-2*t
    else:
        lambda_a = 27.3
        
    return lambda_a
    

def spec_heat(t):
    """ Specific heat of steel [J/kgK] 
        EN 1993-1-2, 3.4.1.2
    """
    
    if t < 600:
        ca = 425+7.73e-1*t - 1.69e-3*t**2+2.22e-6*t**3
    elif t < 735:
        ca = 666 + 13002/(738-t)
    elif t < 900:
        ca = 545+17820/(t-731)
    else:
        ca = 650
    
    return ca

def therm_elongation_stainless(t):
    """ Relative thermal elongation of stainless steel
        EN 1993-1-2, C.3.1
    """
    
    return (16+4.79e-3*t-1.243e-6*t**2)*(t-20)*1e-6
    

def therm_conductivity_stainless(t):
    """ Thermal conductivity of stainless steel [W/mK]
        EN 1993-1-2, C.3.3
    """
    return 14.6+1.27e-2*t
    

def spec_heat_stainless(t):
    """ Specific heat of stainless steel [J/kgK] 
        EN 1993-1-2, C.3.2
    """
    return 450 + 0.280*t-2.91e-4*t**2 + 1.34e-7*t**3
    

def steel_temp(ksh,AmV,tmax,gas_temp=standard_fire,dt=5,Tsteel_ini=20):
    """ Determine temperature of unprotected steel at time 'T'
        starting from t=0 in time intervals 'dT'
    """
    rho_a = 7850 # kg/m3
    
    t = 0
    Tsteel = [Tsteel_ini]
    Time = [0]
    
    if gas_temp == standard_fire:
        aconv = 25
    elif gas_temp == hydrocarbon_fire:
        aconv = 50
    
    while t < tmax:
        """ Assume that gas_temp is a function that takes
            the time in minutes and returns the temperature
            in Celsius.
            
            Alternatively 'gas_temp' could be a table of
            pre-calculated temperatures. Then, dt should match
            the time interval used to produce the table.
            This feature is NOT yet implemented.
        """
        t_gas = gas_temp(t/60)
        
        dTsteel = ksh*AmV/spec_heat(Tsteel[-1])/rho_a*hnet(t_gas,Tsteel[-1],ac=aconv)*dt
        Tsteel.append(Tsteel[-1]+dTsteel)
        t += dt
        Time.append(t/60)
    
    return Tsteel, Time

def stainless_steel_temp(ksh,AmV,tmax,gas_temp=standard_fire,dt=5,Tsteel_ini=20):
    """ Determine temperature of unprotected stainless steel at time 'T'
        starting from t=0 in time intervals 'dT'
    """
    rho_a = 7850 # kg/m3
    
    t = 0
    Tsteel = [Tsteel_ini]
    Time = [0]
    
    if gas_temp == standard_fire:
        aconv = 25
    elif gas_temp == hydrocarbon_fire:
        aconv = 50
    
    while t < tmax:
        """ Assume that gas_temp is a function that takes
            the time in minutes and returns the temperature
            in Celsius.
            
            Alternatively 'gas_temp' could be a table of
            pre-calculated temperatures. Then, dt should match
            the time interval used to produce the table.
            This feature is NOT yet implemented.
        """
        t_gas = gas_temp(t/60)
        dTsteel = ksh*AmV/spec_heat_stainless(Tsteel[-1])/rho_a*hnet(t_gas,Tsteel[-1],eps_m=0.4,ac=aconv)*dt
        Tsteel.append(Tsteel[-1]+dTsteel)
        t += dt
        Time.append(t/60)
    
    return Tsteel, Time

def hnet(t_gas,t_steel,phi=1.0,eps_f=1.0,eps_m=0.7,ac=25):
    """ Net heat flux per unit area [W/m2] """
    s = 5.67e-8 # Stefan-Boltzmann -constant [W/(m2K4)]

    """ heat flux from conductivity """
    hnet_c = ac*(t_gas-t_steel)
    
    """ heat flux from radiation """
    hnet_r = phi*eps_m*eps_f*s*((t_gas+273)**4-(t_steel+273)**4)
    
    return hnet_c+hnet_r


def stress_strain_curves(temp=[400,500,600,700],hardening=False,write=False):
    """ Plots several stree-strain curves """
    import matplotlib.pyplot as plt
    
    strain = np.linspace(0,0.2,200)
    st = []
    
    for idx, t in enumerate(temp):
        st.append([])
        for e in strain:
            st[idx].append(stress(e,t,210000,355,hardening)/355)
     
    if write:
        f = open(r'C:\Users\kmela\OneDrive - TUNI.fi\TRY\LaTeX\figures\luku6\jannitys_venyma_hiiliteras_lujittuminen.dat','w')
        top_row = 'Venymä'
        for t in temp:
            top_row += ' {0:4.0f}C'.format(t)
        f.write(top_row + '\n')
        
        for idx, e in enumerate(strain):
            row = "{0:5.3f} ".format(e)
            for s in st:
                row += ' {0:5.3f}'.format(s[idx])
            row += '\n'
            
            f.write(row)
        
    for stresses in st:
        plt.plot(strain,stresses)
    plt.grid(True)
    plt.xticks([0,0.02,0.04,0.06,0.08,0.1,0.12,0.14,0.16,0.18,0.2])
    plt.yticks([0,0.1,0.2,0.3,0.4,0.5,0.6,0.7,0.8,0.9,1.0])

    return strain, st

def stress_strain_curves_stainless(steel="1.4301",temp=[400,600,700,800],write=False):
    """ Plots several stree-strain curves """
    import matplotlib.pyplot as plt
    
    ssteel = sd.StainlessSteel(steel,'hot_strip')
    strains = []
    st = []
    points = 200
    
    for idx, t in enumerate(temp):
        eps_ut = reduce_property(t,'epsU',steel)
        strains.append(np.linspace(0,eps_ut,points))
        st.append([])
        for e in strains[-1]:            
            st[idx].append(stress_stainless(e,t,ssteel.E,ssteel.fy,ssteel.fu)/ssteel.fy)
     
    if write:
        f = open(r'C:\Users\kmela\OneDrive - TUNI.fi\TRY\LaTeX\figures\luku6\jannitys_venyma_rosteri.dat','w')
        top_row = ''
        for t in temp:
            top_row += 'Venymä {0:4.0f}C '.format(t)
        f.write(top_row + '\n')
        
        for idx in range(points):
            row = ''
            for e, s in zip(strains,st):
                row += "{0:5.3f} {1:5.3f} ".format(e[idx],s[idx])
            row += '\n'
            
            f.write(row)        
        
    for strain, stresses in zip(strains,st):
        plt.plot(strain,stresses)
    plt.grid(True)
    plt.xticks(np.arange(0,0.45,0.05))
    plt.yticks(np.arange(0,1.9,0.1))

    return strains, st

def unprotected_sections(write=False):
    import matplotlib.pyplot as plt
    from sections.steel.ISection import IPE, HEA, HEB, HEM
    from sections.steel.RHS import SHS
    
    
    p1 = IPE(300)
    p2 = HEA(300)
    p3 = HEB(300)
    p4 = HEM(300)
    profiles = [p1,p2,p3,p4]
    
    #p1 = SHS(180,5)
    #p2 = SHS(180,8)
    #p3 = SHS(200,12.5)
    #p3 = SHS(200,20)
    #profiles = [p1,p2,p3]
    Tsteel = []
    Tstainless = []
    
    tmin = 30
    
    for idx, p in enumerate(profiles):
        Tsteel.append([])
        Tstainless.append([])
        ksh = p.shadow_effect()
        AmV = p.section_factor
        
        Tsteel[idx], times = steel_temp(ksh,AmV,tmin*60,gas_temp=hydrocarbon_fire)
        #Tstainless[idx], times = stainless_steel_temp(ksh,AmV,tmin*60,gas_temp=standard_fire)
        Tgas = []
        for t in times:
            Tgas.append(standard_fire(t))
        
        plt.plot(times,Tsteel[idx])
        plt.plot(times,Tgas)
        #plt.plot(times,Tstainless[idx])
        plt.grid(True)
        #plt.xticks(np.arange(0,65,5))
        #plt.yticks(np.arange(0,1050,50))
        plt.xticks(np.arange(0,35,5))
        plt.yticks(np.arange(0,900,50))
    
    if write:
        f = open(r'C:\Users\kmela\OneDrive - TUNI.fi\TRY\LaTeX\figures\luku6\suojaamattomat_IH_hiilivety.dat','w')
        """
        top_row = 'Venymä'
        for t in temp:
            top_row += ' {0:4.0f}C'.format(t)
        f.write(top_row + '\n')
        """
        for idx, t in enumerate(times):
            row = "{0:5.3f} ".format(t)
            for T in Tsteel:
            #for T, Tstain in zip(Tsteel,Tstainless):
                row += ' {0:5.3f}'.format(T[idx])
                #row += ' {0:5.3f}'.format(Tstain[idx])
            row += '\n'
            
            f.write(row)
    
if __name__ == "__main__":
    
    unprotected_sections(write=True)
    #strain,st = stress_strain_curves(write=True)
    #strain,st = stress_strain_curves(temp=[300,350,400],hardening=True,write=True)
    #strain,st = stress_strain_curves_stainless(steel="1.4301",temp=[400,600,700,800],write=True)
    

    
    
    